@@ -65,9 +65,8 @@ eprint(f"Supported platforms: {job_mapping.keys()}")
 class Artifact(NamedTuple):
     job_name: str
     download_name: str
-    output_name: str
+    output_name: Optional[str]
     subdir: str
-    anchor_name: str
 
 # Platform spec provides a specification which is agnostic to Job
 # PlatformSpecs are converted into Artifacts by looking in the jobs-metadata.json file.
@@ -77,14 +76,12 @@ class PlatformSpec(NamedTuple):
 
 source_artifact = Artifact('source-tarball'
                           , 'ghc-{version}-src.tar.xz'
-                          , 'ghc-{version}-src.tar.xz'
-                          , 'ghc-{version}'
-                          , 'ghc{version}-src')
+                          , None
+                          , 'ghc-{version}' )
 test_artifact = Artifact('source-tarball'
                         , 'ghc-{version}-testsuite.tar.xz'
-                        , 'ghc-{version}-testsuite.tar.xz'
-                        , 'ghc-{version}/testsuite'
-                        , 'ghc{version}-testsuite')
+                        , None
+                        , 'ghc-{version}/testsuite' )
 
 def debian(arch, n):
     return linux_platform(arch, "{arch}-linux-deb{n}".format(arch=arch, n=n))
@@ -115,10 +112,9 @@ def linux_platform(arch, opsys):
     return PlatformSpec( opsys, 'ghc-{version}-{arch}-unknown-linux'.format(version="{version}", arch=arch) )
 
 
-base_url = 'https://gitlab.haskell.org/ghc/ghc/-/jobs/{job_id}/artifacts/raw/{artifact_name}'
+base_url = 'https://gitlab.haskell.org/api/v4/projects/1/jobs/{job_id}/artifacts/{artifact_name}'
 
-
-hash_cache = {}
+hash_cache = {} # type: Dict[str, str]
 
 # Download a URL and return its hash
 def download_and_hash(url):
@@ -136,8 +132,6 @@ def download_and_hash(url):
     hash_cache[url] = digest
     return digest
 
-uri_to_anchor_cache=dict()
-
 # Make the metadata for one platform.
 def mk_one_metadata(release_mode, version, job_map, artifact):
     job_id = job_map[artifact.job_name].id
@@ -169,38 +163,31 @@ def mk_one_metadata(release_mode, version, job_map, artifact):
           , "dlSubdir": artifact.subdir.format(version=version)
           , "dlHash" : h }
 
-    # Only add dlOutput if it is inconsistent with the filename inferred from the URL
-    output = artifact.output_name.format(version=version)
-    if Path(urlparse(final_url).path).name != output:
-        res["dlOutput"] = output
-
     eprint(res)
-
-    # add the uri to the anchor name cache so we can lookup an anchor for this uri
-    uri_to_anchor_cache[final_url] = artifact.anchor_name
     return res
 
 # Turns a platform into an Artifact respecting pipeline_type
 # Looks up the right job to use from the .gitlab/jobs-metadata.json file
-def mk_from_platform(pipeline_type, platform):
+def mk_from_platform(release_mode, pipeline_type, platform):
     info = job_mapping[platform.name][pipeline_type]
     eprint(f"From {platform.name} / {pipeline_type} selecting {info['name']}")
+    output_name = None
+    if not release_mode:
+        output_name = "ghc-{version}-{pn}.tar.xz".format(version="{version}", pn=platform.name)
     return Artifact(info['name']
                    , f"{info['jobInfo']['bindistName']}.tar.xz"
-                   , "ghc-{version}-{pn}.tar.xz".format(version="{version}", pn=platform.name)
-                   , platform.subdir
-                   , f"ghc{{version}}-{platform.name}")
+                   , output_name
+                   , platform.subdir)
 
 
 # Generate the new metadata for a specific GHC mode etc
-def mk_new_yaml(release_mode, version, pipeline_type, job_map):
+def mk_new_yaml(release_mode, version, date, pipeline_type, job_map):
     def mk(platform):
         eprint("\n=== " + platform.name + " " + ('=' * (75 - len(platform.name))))
-        return mk_one_metadata(release_mode, version, job_map, mk_from_platform(pipeline_type, platform))
+        return mk_one_metadata(release_mode, version, job_map, mk_from_platform(release_mode, pipeline_type, platform))
 
     ubuntu1804 = mk(ubuntu("18_04"))
     ubuntu2004 = mk(ubuntu("20_04"))
-    ubuntu2204 = mk(ubuntu("22_04"))
     rocky8 = mk(rocky("8"))
     # Here are all the bindists we can distribute
     centos7 = mk(centos(7))
@@ -213,7 +200,7 @@ def mk_new_yaml(release_mode, version, pipeline_type, job_map):
     deb10 = mk(debian("x86_64", 10))
     deb11 = mk(debian("x86_64", 11))
     deb10_arm64 = mk(debian("aarch64", 10))
-    deb9_i386 = mk(debian("i386", 9))
+    deb10_i386 = mk(debian("i386", 10))
 
     source = mk_one_metadata(release_mode, version, job_map, source_artifact)
     test = mk_one_metadata(release_mode, version, job_map, test_artifact)
@@ -228,8 +215,6 @@ def mk_new_yaml(release_mode, version, pipeline_type, job_map):
           , "Linux_Ubuntu" : { "unknown_versioning": ubuntu2004
                              , "( >= 16 && < 18 )": deb9
                              , "( >= 18 && < 19 )": ubuntu1804
-                             , "( >= 19 && < 21 )": ubuntu2004
-                             , "( >= 21 )": ubuntu2204
                              }
           , "Linux_Mint"   : { "< 20": ubuntu1804
                              , ">= 20": ubuntu2004 }
@@ -237,9 +222,7 @@ def mk_new_yaml(release_mode, version, pipeline_type, job_map):
                               , "unknown_versioning" : centos7  }
           , "Linux_Fedora"  : { ">= 33": fedora33
                               , "unknown_versioning": centos7 }
-          , "Linux_RedHat"  : { "< 9": centos7
-                              , ">= 9": fedora33
-                              , "unknown_versioning": fedora33 }
+          , "Linux_RedHat"  : { "unknown_versioning": centos7 }
           #MP: Replace here with Rocky8 when that job is in the pipeline
           , "Linux_UnknownLinux" : { "unknown_versioning": rocky8 }
           , "Darwin" : { "unknown_versioning" : darwin_x86 }
@@ -248,10 +231,10 @@ def mk_new_yaml(release_mode, version, pipeline_type, job_map):
 
           }
 
-    a32 = { "Linux_Debian": { "<10": deb9_i386, "unknown_versioning": deb9_i386 }
-          , "Linux_Ubuntu": { "unknown_versioning": deb9_i386 }
-          , "Linux_Mint" : { "unknown_versioning": deb9_i386 }
-          , "Linux_UnknownLinux" : { "unknown_versioning": deb9_i386 }
+    a32 = { "Linux_Debian": { "unknown_versioning": deb10_i386 }
+          , "Linux_Ubuntu": { "unknown_versioning": deb10_i386 }
+          , "Linux_Mint" : { "unknown_versioning": deb10_i386 }
+          , "Linux_UnknownLinux" : { "unknown_versioning": deb10_i386 }
           }
 
     arm64 = { "Linux_UnknownLinux": { "unknown_versioning": deb10_arm64 }
@@ -268,7 +251,14 @@ def mk_new_yaml(release_mode, version, pipeline_type, job_map):
     else:
         change_log =  "https://gitlab.haskell.org"
 
-    return { "viTags": ["Latest", "TODO_base_version"]
+    if release_mode:
+        tags =  ["Latest", "TODO_base_version"]
+    else:
+        tags = ["LatestNightly"]
+
+
+    return { "viTags": tags
+        , "viReleaseDay": date
         # Check that this link exists
         , "viChangeLog": change_log
         , "viSourceDL": source
@@ -280,19 +270,13 @@ def mk_new_yaml(release_mode, version, pipeline_type, job_map):
         }
 
 
-def mk_dumper(version):
-  class CustomAliasDumper(yaml.Dumper):
-      def __init__(self, *args, **kwargs):
-          super().__init__(*args, **kwargs)
+def setNightlyTags(ghcup_metadata):
+    for version in ghcup_metadata['ghcupDownloads']['GHC']:
+        if "LatestNightly" in ghcup_metadata['ghcupDownloads']['GHC'][version]["viTags"]:
+            ghcup_metadata['ghcupDownloads']['GHC'][version]["viTags"].remove("LatestNightly")
+            ghcup_metadata['ghcupDownloads']['GHC'][version]["viTags"].append("Nightly")
 
-      def generate_anchor(self, node):
-          if isinstance(node, yaml.MappingNode):
-            node_dict = { k.value : v.value for (k,v) in node.value }
-            if 'dlUri' in node_dict:
-              return uri_to_anchor_cache[node_dict['dlUri']].format(version=version.replace('.',''))
-          return super().generate_anchor(node)
 
-  return CustomAliasDumper
 
 
 def main() -> None:
@@ -305,6 +289,7 @@ def main() -> None:
     parser.add_argument('--fragment', action='store_true', help='Output the generated fragment rather than whole modified file')
     # TODO: We could work out the --version from the project-version CI job.
     parser.add_argument('--version', required=True, type=str, help='Version of the GHC compiler')
+    parser.add_argument('--date', required=True, type=str, help='Date of the compiler release')
     args = parser.parse_args()
 
     project = gl.projects.get(1, lazy=True)
@@ -325,17 +310,21 @@ def main() -> None:
     eprint(f"Pipeline Type: {pipeline_type}")
 
 
-    new_yaml = mk_new_yaml(args.release_mode, args.version, pipeline_type, job_map)
+    new_yaml = mk_new_yaml(args.release_mode, args.version, args.date, pipeline_type, job_map)
     if args.fragment:
-        print(yaml.dump({ args.version : new_yaml }, Dumper=mk_dumper(args.version)))
+        print(yaml.dump({ args.version : new_yaml }))
 
     else:
         with open(args.metadata, 'r') as file:
             ghcup_metadata = yaml.safe_load(file)
+            if  args.version in ghcup_metadata['ghcupDownloads']['GHC']:
+                raise RuntimeError("Refusing to override existing version in metadata")
+            setNightlyTags(ghcup_metadata)
             ghcup_metadata['ghcupDownloads']['GHC'][args.version] = new_yaml
             print(yaml.dump(ghcup_metadata))
 
 
+
 if __name__ == '__main__':
     main()
 
