@@ -1,10 +1,6 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE RecordWildCards #-}
-{-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TupleSections #-}
 {-# LANGUAGE LambdaCase #-}
-{-# LANGUAGE RankNTypes #-}
 
 -- | Interacting with the iserv interpreter, whether it is running on an
 -- external process or in the current process.
@@ -25,19 +21,20 @@ module GHC.Runtime.Interpreter
   , mkCostCentres
   , costCentreStackInfo
   , newBreakArray
+  , newModuleName
   , storeBreakpoint
   , breakpointStatus
   , getBreakpointVar
   , getClosure
   , getModBreaks
   , seqHValue
+  , evalBreakInfo
   , interpreterDynamic
   , interpreterProfiled
 
   -- * The object-code linker
   , initObjLinker
   , lookupSymbol
-  , lookupSymbolInDLL
   , lookupClosure
   , loadDLL
   , loadArchive
@@ -84,7 +81,6 @@ import GHC.Linker.Types
 import GHC.Data.Maybe
 import GHC.Data.FastString
 
-import GHC.Types.Unique
 import GHC.Types.SrcLoc
 import GHC.Types.Unique.FM
 import GHC.Types.Basic
@@ -93,7 +89,6 @@ import GHC.Utils.Panic
 import GHC.Utils.Exception as Ex
 import GHC.Utils.Outputable(brackets, ppr, showSDocUnsafe)
 import GHC.Utils.Fingerprint
-import GHC.Utils.Misc
 
 import GHC.Unit.Module
 import GHC.Unit.Module.ModIface
@@ -110,9 +105,7 @@ import Control.Monad
 import Control.Monad.IO.Class
 import Control.Monad.Catch as MC (mask)
 import Data.Binary
-import Data.Binary.Put
 import Data.ByteString (ByteString)
-import qualified Data.ByteString.Lazy as LB
 import Data.Array ((!))
 import Data.IORef
 import Foreign hiding (void)
@@ -120,7 +113,6 @@ import qualified GHC.Exts.Heap as Heap
 import GHC.Stack.CCS (CostCentre,CostCentreStack)
 import System.Directory
 import System.Process
-import GHC.Conc (pseq, par)
 
 {- Note [Remote GHCi]
    ~~~~~~~~~~~~~~~~~~
@@ -159,22 +151,22 @@ The main pieces are:
   - implementation of Template Haskell (GHCi.TH)
   - a few other things needed to run interpreted code
 
-- top-level iserv directory, containing the code for the external
-  server. This is a fairly simple wrapper, most of the functionality
+- top-level iserv directory, containing the codefor the external
+  server.  This is a fairly simple wrapper, most of the functionality
   is provided by modules in libraries/ghci.
 
 - This module which provides the interface to the server used
   by the rest of GHC.
 
-GHC works with and without -fexternal-interpreter. With the flag, all
-interpreted code is run by the iserv binary. Without the flag,
+GHC works with and without -fexternal-interpreter.  With the flag, all
+interpreted code is run by the iserv binary.  Without the flag,
 interpreted code is run in the same process as GHC.
 
 Things that do not work with -fexternal-interpreter
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
 dynCompileExpr cannot work, because we have no way to run code of an
-unknown type in the remote process. This API fails with an error
+unknown type in the remote process.  This API fails with an error
 message if it is used with -fexternal-interpreter.
 
 Other Notes on Remote GHCi
@@ -314,7 +306,7 @@ handleEvalStatus
   -> IO (EvalStatus_ [ForeignHValue] [HValueRef])
 handleEvalStatus interp status =
   case status of
-    EvalBreak a b c d e f -> return (EvalBreak a b c d e f)
+    EvalBreak a b c d -> return (EvalBreak a b c d)
     EvalComplete alloc res ->
       EvalComplete alloc <$> addFinalizer res
  where
@@ -353,19 +345,7 @@ mkCostCentres interp mod ccs =
 -- | Create a set of BCOs that may be mutually recursive.
 createBCOs :: Interp -> [ResolvedBCO] -> IO [HValueRef]
 createBCOs interp rbcos = do
-  -- Serializing ResolvedBCO is expensive, so we do it in parallel
-  interpCmd interp (CreateBCOs puts)
- where
-  puts = parMap doChunk (chunkList 100 rbcos)
-
-  -- make sure we force the whole lazy ByteString
-  doChunk c = pseq (LB.length bs) bs
-    where bs = runPut (put c)
-
-  -- We don't have the parallel package, so roll our own simple parMap
-  parMap _ [] = []
-  parMap f (x:xs) = fx `par` (fxs `pseq` (fx : fxs))
-    where fx = f x; fxs = parMap f xs
+  interpCmd interp (CreateBCOs rbcos)
 
 addSptEntry :: Interp -> Fingerprint -> ForeignHValue -> IO ()
 addSptEntry interp fpr ref =
@@ -381,6 +361,10 @@ newBreakArray interp size = do
   breakArray <- interpCmd interp (NewBreakArray size)
   mkFinalizedHValue interp breakArray
 
+newModuleName :: Interp -> ModuleName -> IO (RemotePtr ModuleName)
+newModuleName interp mod_name =
+  castRemotePtr <$> interpCmd interp (NewBreakModule (moduleNameString mod_name))
+
 storeBreakpoint :: Interp -> ForeignRef BreakArray -> Int -> Int -> IO ()
 storeBreakpoint interp ref ix cnt = do                               -- #19157
   withForeignRef ref $ \breakarray ->
@@ -410,20 +394,24 @@ seqHValue interp unit_env ref =
     status <- interpCmd interp (Seq hval)
     handleSeqHValueStatus interp unit_env status
 
+evalBreakInfo :: HomePackageTable -> EvalBreakpoint -> BreakInfo
+evalBreakInfo hpt (EvalBreakpoint ix mod_name) =
+  BreakInfo modl ix
+  where
+    modl = mi_module $
+           hm_iface $
+           expectJust "evalBreakInfo" $
+           lookupHpt hpt (mkModuleName mod_name)
+
 -- | Process the result of a Seq or ResumeSeq message.             #2950
 handleSeqHValueStatus :: Interp -> UnitEnv -> EvalStatus () -> IO (EvalResult ())
 handleSeqHValueStatus interp unit_env eval_status =
   case eval_status of
-    (EvalBreak is_exception _ ix mod_uniq resume_ctxt _) -> do
+    (EvalBreak _ maybe_break resume_ctxt _) -> do
       -- A breakpoint was hit; inform the user and tell them
       -- which breakpoint was hit.
       resume_ctxt_fhv <- liftIO $ mkFinalizedHValue interp resume_ctxt
-      let hmi = expectJust "handleRunStatus" $
-                  lookupHptDirectly (ue_hpt unit_env)
-                    (mkUniqueGrimily mod_uniq)
-          modl = mi_module (hm_iface hmi)
-          bp | is_exception = Nothing
-             | otherwise = Just (BreakInfo modl ix)
+      let bp = evalBreakInfo (ue_hpt unit_env) <$> maybe_break
           sdocBpLoc = brackets . ppr . getSeqBpSpan
       putStrLn ("*** Ignoring breakpoint " ++
             (showSDocUnsafe $ sdocBpLoc bp))
@@ -463,7 +451,7 @@ lookupSymbol interp str = case interpInstance interp of
       -- making cross-process LookupSymbol calls, so I added a GHC-side
       -- cache which sped things up quite a lot.  We have to be careful
       -- to purge this cache when unloading code though.
-      cache <- readMVar (interpLookupSymbolCache interp)
+      cache <- readMVar (instLookupSymbolCache inst)
       case lookupUFM cache str of
         Just p -> return (Just p)
         Nothing -> do
@@ -474,72 +462,35 @@ lookupSymbol interp str = case interpInstance interp of
             Just r -> do
               let p        = fromRemotePtr r
                   cache'   = addToUFM cache str p
-              modifyMVar_ (interpLookupSymbolCache interp) (const (pure cache'))
+              modifyMVar_ (instLookupSymbolCache inst) (const (pure cache'))
               return (Just p)
 
     ExtJS {} -> pprPanic "lookupSymbol not supported by the JS interpreter" (ppr str)
 
-lookupSymbolInDLL :: Interp -> RemotePtr LoadedDLL -> FastString -> IO (Maybe (Ptr ()))
-lookupSymbolInDLL interp dll str = withSymbolCache interp str $
-  case interpInstance interp of
-#if defined(HAVE_INTERNAL_INTERPRETER)
-    InternalInterp -> fmap fromRemotePtr <$> run (LookupSymbolInDLL dll (unpackFS str))
-#endif
-    ExternalInterp ext -> case ext of
-      ExtIServ i -> withIServ i $ \inst -> fmap fromRemotePtr <$> do
-        uninterruptibleMask_ $
-          sendMessage inst (LookupSymbolInDLL dll (unpackFS str))
-      ExtJS {} -> pprPanic "lookupSymbol not supported by the JS interpreter" (ppr str)
-
 lookupClosure :: Interp -> String -> IO (Maybe HValueRef)
 lookupClosure interp str =
   interpCmd interp (LookupClosure str)
 
--- | 'withSymbolCache' tries to find a symbol in the 'interpLookupSymbolCache'
--- which maps symbols to the address where they are loaded.
--- When there's a cache hit we simply return the cached address, when there is
--- a miss we run the action which determines the symbol's address and populate
--- the cache with the answer.
-withSymbolCache :: Interp
-                -> FastString
-                -- ^ The symbol we are looking up in the cache
-                -> IO (Maybe (Ptr ()))
-                -- ^ An action which determines the address of the symbol we
-                -- are looking up in the cache, which is run if there is a
-                -- cache miss. The result will be cached.
-                -> IO (Maybe (Ptr ()))
-withSymbolCache interp str determine_addr = do
-
-  -- Profiling of GHCi showed a lot of time and allocation spent
-  -- making cross-process LookupSymbol calls, so I added a GHC-side
-  -- cache which sped things up quite a lot. We have to be careful
-  -- to purge this cache when unloading code though.
-  --
-  -- The analysis in #23415 further showed this cache should also benefit the
-  -- internal interpreter's loading times, and needn't be used by the external
-  -- interpreter only.
-  cache <- readMVar (interpLookupSymbolCache interp)
-  case lookupUFM cache str of
-    Just p -> return (Just p)
-    Nothing -> do
-
-      maddr <- determine_addr
-      case maddr of
-        Nothing -> return Nothing
-        Just p -> do
-          let upd_cache cache' = addToUFM cache' str p
-          modifyMVar_ (interpLookupSymbolCache interp) (pure . upd_cache)
-          return (Just p)
-
 purgeLookupSymbolCache :: Interp -> IO ()
-purgeLookupSymbolCache interp = modifyMVar_ (interpLookupSymbolCache interp) (const (pure emptyUFM))
+purgeLookupSymbolCache interp = case interpInstance interp of
+#if defined(HAVE_INTERNAL_INTERPRETER)
+  InternalInterp -> pure ()
+#endif
+  ExternalInterp ext -> withExtInterpMaybe ext $ \case
+    Nothing   -> pure () -- interpreter stopped, nothing to do
+    Just inst -> modifyMVar_ (instLookupSymbolCache inst) (const (pure emptyUFM))
 
 -- | loadDLL loads a dynamic library using the OS's native linker
 -- (i.e. dlopen() on Unix, LoadLibrary() on Windows).  It takes either
 -- an absolute pathname to the file, or a relative filename
 -- (e.g. "libfoo.so" or "foo.dll").  In the latter case, loadDLL
 -- searches the standard locations for the appropriate library.
-loadDLL :: Interp -> String -> IO (Either String (RemotePtr LoadedDLL))
+--
+-- Returns:
+--
+-- Nothing      => success
+-- Just err_msg => failure
+loadDLL :: Interp -> String -> IO (Maybe String)
 loadDLL interp str = interpCmd interp (LoadDLL str)
 
 loadArchive :: Interp -> String -> IO ()
@@ -598,9 +549,11 @@ spawnIServ conf = do
                   }
 
   pending_frees <- newMVar []
+  lookup_cache  <- newMVar emptyUFM
   let inst = ExtInterpInstance
         { instProcess           = process
         , instPendingFrees      = pending_frees
+        , instLookupSymbolCache = lookup_cache
         , instExtra             = ()
         }
   pure inst
