@@ -7,7 +7,7 @@ module Settings.Default (
 
     -- * Default command line arguments for various builders
     SourceArgs (..), sourceArgs, defaultBuilderArgs, defaultPackageArgs,
-    defaultArgs,
+    defaultExtraArgs, defaultHaddockExtraArgs,
 
     -- * Default build flavour and BigNum backend
     defaultFlavour, defaultBignumBackend
@@ -17,7 +17,6 @@ import qualified Data.Set as Set
 
 import qualified Hadrian.Builder.Sphinx
 import qualified Hadrian.Builder.Tar
-import Hadrian.Haskell.Cabal.Type
 
 import CommandLine
 import Expression
@@ -27,6 +26,7 @@ import Oracles.Setting
 import Packages
 import Settings.Builders.Alex
 import Settings.Builders.DeriveConstants
+import Settings.Builders.GenApply
 import Settings.Builders.Cabal
 import Settings.Builders.Cc
 import Settings.Builders.Configure
@@ -64,9 +64,17 @@ defaultBignumBackend = "gmp"
 -- These packages are things needed to do the build.. so they are only built by
 -- boot compiler, with global package database. By default we will only build these
 -- packages in StageBoot so if you also need to distribute anything here then add
--- it to `stage0packages` or `stage1packages` as appropiate.
+-- it to `stage0packages` or `stage1packages` as appropriate.
 stageBootPackages :: Action [Package]
-stageBootPackages = return [lintersCommon, lintCommitMsg, lintSubmoduleRefs, lintWhitespace, lintNotes, hsc2hs, compareSizes, deriveConstants, genapply, genprimopcode, unlit ]
+stageBootPackages = return
+  [ lintersCommon, lintCommitMsg, lintSubmoduleRefs, lintWhitespace, lintNotes
+  , hsc2hs
+  , compareSizes
+  , deriveConstants
+  , genapply
+  , genprimopcode
+  , unlit
+  ]
 
 -- | Packages built in 'Stage0' by default. You can change this in "UserSettings".
 stage0Packages :: Action [Package]
@@ -86,7 +94,9 @@ stage0Packages = do
              , runGhc
              , ghcBoot
              , ghcBootTh
+             , ghcPlatform
              , ghcHeap
+             , ghcToolchain
              , ghci
              , ghcPkg
              , haddock
@@ -94,7 +104,9 @@ stage0Packages = do
              , hpc
              , hpcBin
              , mtl
+             , osString
              , parsec
+             , semaphoreCompat
              , time
              , templateHaskell
              , text
@@ -105,7 +117,6 @@ stage0Packages = do
              ]
           ++ [ terminfo | not windowsHost, not cross ]
           ++ [ timeout  | windowsHost                ]
-          ++ [ touchy   | windowsHost                ]
 
 -- | Packages built in 'Stage1' by default. You can change this in "UserSettings".
 stage1Packages :: Action [Package]
@@ -134,6 +145,8 @@ stage1Packages = do
         , ghc
         , ghcBignum
         , ghcCompact
+        , ghcExperimental
+        , ghcInternal
         , ghcPkg
         , ghcPrim
         , haskeline
@@ -142,6 +155,7 @@ stage1Packages = do
         , integerGmp
         , pretty
         , rts
+        , semaphoreCompat
         , stm
         , unlit
         , xhtml
@@ -151,13 +165,12 @@ stage1Packages = do
         [ haddock
         , hpcBin
         , iserv
-        , libiserv
         , runGhc
+        , ghcToolchainBin
         ]
       , when (winTarget && not cross)
-        [ touchy
-         -- See Note [Hadrian's ghci-wrapper package]
-        , ghciWrapper
+        [ -- See Note [Hadrian's ghci-wrapper package]
+          ghciWrapper
         ]
       ]
 
@@ -167,7 +180,7 @@ stage2Packages = stage1Packages
 
 -- | Packages that are built only for the testsuite.
 testsuitePackages :: Action [Package]
-testsuitePackages = return ([ timeout | windowsHost ] ++ [ checkPpr, checkExact, countDeps, ghcConfig ])
+testsuitePackages = return ([ timeout | windowsHost ] ++ [ checkPpr, checkExact, countDeps, lintCodes, ghcConfig, dumpDecls ])
 
 -- | Default build ways for library packages:
 -- * We always build 'vanilla' way.
@@ -209,7 +222,6 @@ data SourceArgs = SourceArgs
 sourceArgs :: SourceArgs -> Args
 sourceArgs SourceArgs {..} = builder Ghc ? mconcat
     [ hsDefault
-    , getContextData hcOpts
     -- `compiler` is also a library but the specific arguments that we want
     -- to apply to that are given by the hsCompiler option. `ghc` is an
     -- executable so we don't have to exclude that.
@@ -218,11 +230,14 @@ sourceArgs SourceArgs {..} = builder Ghc ? mconcat
     , package ghc      ? hsGhc ]
 
 -- | All default command line arguments.
-defaultArgs :: Args
-defaultArgs = mconcat
-    [ defaultBuilderArgs
-    , sourceArgs defaultSourceArgs
-    , defaultPackageArgs ]
+defaultExtraArgs :: Args
+defaultExtraArgs =
+  mconcat [ sourceArgs defaultSourceArgs, defaultHaddockExtraArgs ]
+
+defaultHaddockExtraArgs :: Args
+defaultHaddockExtraArgs = builder (Haddock BuildPackage) ?
+  mconcat [ arg "--hyperlinked-source", arg "--hoogle", arg "--quickjump" ]
+
 
 -- | Default source arguments, e.g. optimisation settings.
 defaultSourceArgs :: SourceArgs
@@ -240,7 +255,7 @@ defaultSourceArgs = SourceArgs
 defaultFlavour :: Flavour
 defaultFlavour = Flavour
     { name               = "default"
-    , args               = defaultArgs
+    , extraArgs          = defaultExtraArgs
     , packages           = defaultPackages
     , bignumBackend      = defaultBignumBackend
     , bignumCheck        = False
@@ -274,6 +289,7 @@ defaultBuilderArgs = mconcat
     , ccBuilderArgs
     , configureBuilderArgs
     , deriveConstantsBuilderArgs
+    , genapplyBuilderArgs
     , genPrimopCodeBuilderArgs
     , ghcBuilderArgs
     , ghcPkgBuilderArgs
