@@ -4,47 +4,26 @@ set -Eeuo pipefail
 
 source "$CI_PROJECT_DIR/.gitlab/common.sh"
 
-export GHCUP_INSTALL_BASE_PREFIX="$CI_PROJECT_DIR/toolchain"
+if [[ "$(uname)" == "Linux" ]]; then
+    export PATH="/opt/ghc/${GHC_VERSION}/bin:${PATH}"
+# Not all runners use ci-images, so ghcup is used.
+else
+    . "$CI_PROJECT_DIR/.gitlab/ghcup.sh"
+fi
+
 export CABAL_DIR="$CI_PROJECT_DIR/cabal"
 
 case "$(uname)" in
     MSYS_*|MINGW*)
         export CABAL_DIR="$(cygpath -w "$CABAL_DIR")"
-        GHCUP_BINDIR="${GHCUP_INSTALL_BASE_PREFIX}/ghcup/bin"
         EXE_EXT=".exe"
         ;;
     *)
-        GHCUP_BINDIR="${GHCUP_INSTALL_BASE_PREFIX}/.ghcup/bin"
         EXE_EXT=""
         ;;
 esac
 
 mkdir -p "$CABAL_DIR"
-mkdir -p "$GHCUP_BINDIR"
-export PATH="$GHCUP_BINDIR:$PATH"
-
-export BOOTSTRAP_HASKELL_NONINTERACTIVE=1
-export BOOTSTRAP_HASKELL_GHC_VERSION=$GHC_VERSION
-export BOOTSTRAP_HASKELL_CABAL_VERSION=$CABAL_INSTALL_VERSION
-export BOOTSTRAP_HASKELL_ADJUST_CABAL_CONFIG=yes
-
-# for some reason the subshell doesn't pick up the arm64 environment on darwin
-# and starts installing x86_64 GHC
-case "$(uname -s)" in
-    "Darwin"|"darwin")
-        case "$(/usr/bin/arch)" in
-            aarch64|arm64|armv8l)
-                curl --proto '=https' --tlsv1.2 -sSf https://get-ghcup.haskell.org | arch -arm64 /bin/bash
-                ;;
-            *)
-                curl --proto '=https' --tlsv1.2 -sSf https://get-ghcup.haskell.org | sh
-                ;;
-        esac
-        ;;
-    *)
-        curl --proto '=https' --tlsv1.2 -sSf https://get-ghcup.haskell.org | sh
-        ;;
-esac
 
 # https://github.com/haskell/cabal/issues/7313#issuecomment-811851884
 # and
@@ -56,13 +35,13 @@ if [ "$(getconf LONG_BIT)" = "32" -o "${PLATFORM:=xxx}" = "x86_64-linux-centos7"
 fi
 
 args=(
-    -w "ghc-$GHC_VERSION"
     --disable-profiling
     --enable-executable-stripping
     --project-file=cabal.project.release
     ${ADD_CABAL_ARGS}
 )
 
+run cabal update hackage.haskell.org,HEAD
 run cabal v2-build ${args[@]} cabal-install
 
 mkdir "$CI_PROJECT_DIR/out"
