@@ -29,16 +29,17 @@ import Distribution.Simple.Program.Types
 import Distribution.Simple.Program.Db
 import Distribution.Simple.Program
 import Distribution.System (OS(Windows,Linux,OSX), Arch(JavaScript), buildOS, buildArch)
-import Distribution.Simple.Utils
-    ( withFileContents, withTempDirectory, tryFindPackageDesc )
 import Distribution.Simple.Configure
     ( getPersistBuildConfig )
+import Distribution.Simple.Utils
+    ( withFileContents, tryFindPackageDesc )
 import Distribution.Version
 import Distribution.Package
 import Distribution.Parsec (eitherParsec)
 import Distribution.Types.UnqualComponentName
 import Distribution.Types.LocalBuildInfo
 import Distribution.PackageDescription
+import Test.Utils.TempTestDir (withTestDir)
 import Distribution.Verbosity (normal)
 
 import Distribution.Compat.Stack
@@ -52,23 +53,23 @@ import Control.Monad (unless, when, void, forM_, liftM2, liftM4)
 import Control.Monad.Trans.Reader (withReaderT, runReaderT)
 import Control.Monad.IO.Class (MonadIO (..))
 import qualified Crypto.Hash.SHA256 as SHA256
-import qualified Data.ByteString.Base64 as Base64
+import qualified Data.ByteString.Base16 as Base16
 import qualified Data.ByteString.Char8 as C
 import Data.List (isInfixOf, stripPrefix, isPrefixOf, intercalate)
 import Data.List.NonEmpty (NonEmpty (..))
 import qualified Data.List.NonEmpty as NE
 import Data.Maybe (mapMaybe, fromMaybe)
 import System.Exit (ExitCode (..))
-import System.FilePath ((</>), takeExtensions, takeDrive, takeDirectory, normalise, splitPath, joinPath, splitFileName, (<.>), dropTrailingPathSeparator)
+import System.FilePath
 import Control.Concurrent (threadDelay)
 import qualified Data.Char as Char
-import System.Directory (getTemporaryDirectory, getCurrentDirectory, canonicalizePath, copyFile, copyFile, doesDirectoryExist, doesFileExist, createDirectoryIfMissing, getDirectoryContents, listDirectory)
+import System.Directory
 import Control.Retry (exponentialBackoff, limitRetriesByCumulativeDelay)
 import Network.Wait (waitTcpVerbose)
+import System.Environment
 
 #ifndef mingw32_HOST_OS
 import Control.Monad.Catch ( bracket_ )
-import System.Directory    ( removeFile )
 import System.Posix.Files  ( createSymbolicLink )
 import System.Posix.Resource
 #endif
@@ -112,6 +113,16 @@ withDirectory f = withReaderT
 withEnv :: [(String, Maybe String)] -> TestM a -> TestM a
 withEnv e = withReaderT (\env -> env { testEnvironment = testEnvironment env ++ e })
 
+-- | Prepend a directory to the PATH
+addToPath :: FilePath -> TestM a -> TestM a
+addToPath exe_dir action = do
+  env <- getTestEnv
+  path <- liftIO $ getEnv "PATH"
+  let newpath = exe_dir ++ [searchPathSeparator] ++ path
+  let new_env = (("PATH", Just newpath) : (testEnvironment env))
+  withEnv new_env action
+
+
 -- HACK please don't use me
 withEnvFilter :: (String -> Bool) -> TestM a -> TestM a
 withEnvFilter p = withReaderT (\env -> env { testEnvironment = filter (p . fst) (testEnvironment env) })
@@ -278,9 +289,6 @@ cabalG' global_args cmd args = cabalGArgs global_args cmd args Nothing
 cabalGArgs :: [String] -> String -> [String] -> Maybe String -> TestM Result
 cabalGArgs global_args cmd args input = do
     env <- getTestEnv
-    -- Freeze writes out cabal.config to source directory, this is not
-    -- overwritable
-    when (cmd == "v1-freeze") requireHasSourceCopy
     let extra_args
           | cmd `elem`
               [ "v1-update"
@@ -298,17 +306,17 @@ cabalGArgs global_args cmd args input = do
           = [ ]
 
           -- new-build commands are affected by testCabalProjectFile
-          | cmd == "v2-sdist"
-          = [ "--project-file", testCabalProjectFile env ]
+          | cmd `elem` ["v2-sdist", "path"]
+          = [ "--project-file=" ++ fp | Just fp <- [testCabalProjectFile env] ]
 
           | cmd == "v2-clean"
-          = [ "--builddir", testDistDir env
-            , "--project-file", testCabalProjectFile env ]
+          = [ "--builddir", testDistDir env ]
+            ++ [ "--project-file=" ++ fp | Just fp <- [testCabalProjectFile env] ]
 
           | "v2-" `isPrefixOf` cmd
           = [ "--builddir", testDistDir env
-            , "--project-file", testCabalProjectFile env
             , "-j1" ]
+            ++ [ "--project-file=" ++ fp | Just fp <- [testCabalProjectFile env] ]
 
           | otherwise
           = [ "--builddir", testDistDir env ] ++
@@ -331,7 +339,7 @@ cabal_raw' cabal_args input = runProgramM cabalProgram cabal_args input
 
 withProjectFile :: FilePath -> TestM a -> TestM a
 withProjectFile fp m =
-    withReaderT (\env -> env { testCabalProjectFile = fp }) m
+    withReaderT (\env -> env { testCabalProjectFile = Just fp }) m
 
 -- | Assuming we've successfully configured a new-build project,
 -- read out the plan metadata so that we can use it to do other
@@ -356,15 +364,21 @@ runPlanExe pkg_name cname args = void $ runPlanExe' pkg_name cname args
 runPlanExe' :: String {- package name -} -> String {- component name -}
             -> [String] -> TestM Result
 runPlanExe' pkg_name cname args = do
+    exePath <- planExePath pkg_name cname
+    defaultRecordMode RecordAll $ do
+    recordHeader [pkg_name, cname]
+    runM exePath args Nothing
+
+planExePath :: String {- package name -} -> String {- component name -}
+            -> TestM FilePath
+planExePath pkg_name cname = do
     Just plan <- testPlan `fmap` getTestEnv
     let distDirOrBinFile = planDistDir plan (mkPackageName pkg_name)
                                (CExeName (mkUnqualComponentName cname))
         exePath = case distDirOrBinFile of
           DistDir dist_dir -> dist_dir </> "build" </> cname </> cname
           BinFile bin_file -> bin_file
-    defaultRecordMode RecordAll $ do
-    recordHeader [pkg_name, cname]
-    runM exePath args Nothing
+    return exePath
 
 ------------------------------------------------------------------------
 -- * Running ghc-pkg
@@ -649,26 +663,6 @@ withRemoteRepo repoDir m = do
             runReaderT m (env { testHaveRepo = True }))
 
 
-------------------------------------------------------------------------
--- * Subprocess run results
-
-requireSuccess :: Result -> TestM Result
-requireSuccess r@Result { resultCommand = cmd
-                        , resultExitCode = exitCode
-                        , resultOutput = output } = withFrozenCallStack $ do
-    env <- getTestEnv
-    when (exitCode /= ExitSuccess && not (testShouldFail env)) $
-        assertFailure $ "Command " ++ cmd ++ " failed.\n" ++
-        "Output:\n" ++ output ++ "\n"
-    when (exitCode == ExitSuccess && testShouldFail env) $
-        assertFailure $ "Command " ++ cmd ++ " succeeded.\n" ++
-        "Output:\n" ++ output ++ "\n"
-    return r
-
-initWorkDir :: TestM ()
-initWorkDir = do
-    env <- getTestEnv
-    liftIO $ createDirectoryIfMissing True (testWorkDir env)
 
 -- | Record a header to help identify the output to the expect
 -- log.  Unlike the 'recordLog', we don't record all arguments;
@@ -680,46 +674,21 @@ recordHeader args = do
     env <- getTestEnv
     let mode = testRecordMode env
         str_header = "# " ++ intercalate " " args ++ "\n"
-        header = C.pack (testRecordNormalizer env str_header)
+        rec_header = C.pack str_header
     case mode of
         DoNotRecord -> return ()
         _ -> do
             initWorkDir
             liftIO $ putStr str_header
-            liftIO $ C.appendFile (testWorkDir env </> "test.log") header
-            liftIO $ C.appendFile (testActualFile env) header
+            liftIO $ C.appendFile (testWorkDir env </> "test.log") rec_header
+            liftIO $ C.appendFile (testActualFile env) rec_header
 
-recordLog :: Result -> TestM ()
-recordLog res = do
-    env <- getTestEnv
-    let mode = testRecordMode env
-    initWorkDir
-    liftIO $ C.appendFile (testWorkDir env </> "test.log")
-                         (C.pack $ "+ " ++ resultCommand res ++ "\n"
-                            ++ resultOutput res ++ "\n\n")
-    liftIO . C.appendFile (testActualFile env) . C.pack . testRecordNormalizer env $
-        case mode of
-            RecordAll    -> unlines (lines (resultOutput res))
-            RecordMarked -> getMarkedOutput (resultOutput res)
-            DoNotRecord  -> ""
-
-getMarkedOutput :: String -> String -- trailing newline
-getMarkedOutput out = unlines (go (lines out) False)
-  where
-    go [] _ = []
-    go (x:xs) True
-        | "-----END CABAL OUTPUT-----"   `isPrefixOf` x
-                    =     go xs False
-        | otherwise = x : go xs True
-    go (x:xs) False
-        -- NB: Windows has extra goo at the end
-        | "-----BEGIN CABAL OUTPUT-----" `isPrefixOf` x
-                    = go xs True
-        | otherwise = go xs False
 
 ------------------------------------------------------------------------
 -- * Test helpers
 
+------------------------------------------------------------------------
+-- * Subprocess run results
 assertFailure :: WithCallStack (String -> m ())
 assertFailure msg = withFrozenCallStack $ error msg
 
@@ -788,10 +757,6 @@ recordMode mode = withReaderT (\env -> env {
     testRecordUserMode = Just mode
     })
 
-recordNormalizer :: (String -> String) -> TestM a -> TestM a
-recordNormalizer f =
-    withReaderT (\env -> env { testRecordNormalizer = testRecordNormalizer env . f })
-
 assertOutputContains :: MonadIO m => WithCallStack (String -> Result -> m ())
 assertOutputContains needle result =
     withFrozenCallStack $
@@ -842,8 +807,7 @@ getScriptCacheDirectory :: FilePath -> TestM FilePath
 getScriptCacheDirectory script = do
     cabalDir <- testCabalDir `fmap` getTestEnv
     hashinput <- liftIO $ canonicalizePath script
-    let hash = map (\c -> if c == '/' then '%' else c) . take 26
-             . C.unpack . Base64.encode . SHA256.hash . C.pack $ hashinput
+    let hash = C.unpack . Base16.encode . C.take 26 . SHA256.hash . C.pack $ hashinput
     return $ cabalDir </> "script-builds" </> hash
 
 ------------------------------------------------------------------------
@@ -981,8 +945,7 @@ expectBrokenIf True ticket m = expectBroken ticket m
 expectBrokenUnless :: Bool -> Int -> TestM a -> TestM ()
 expectBrokenUnless b = expectBrokenIf (not b)
 
-------------------------------------------------------------------------
--- * Miscellaneous
+-- * Programs
 
 git :: String -> [String] -> TestM ()
 git cmd args = void $ git' cmd args
@@ -1016,41 +979,6 @@ python3' args = do
     recordHeader ["python3"]
     runProgramM python3Program args Nothing
 
--- | If a test needs to modify or write out source files, it's
--- necessary to make a hermetic copy of the source files to operate
--- on.  This function arranges for this to be done.
---
--- This requires the test repository to be a Git checkout, because
--- we use the Git metadata to figure out what files to copy into the
--- hermetic copy.
---
--- Also see 'withSourceCopyDir'.
-withSourceCopy :: TestM a -> TestM a
-withSourceCopy m = do
-    env <- getTestEnv
-    let cwd  = testCurrentDir env
-        dest = testSourceCopyDir env
-    r <- git' "ls-files" ["--cached", "--modified"]
-    forM_ (lines (resultOutput r)) $ \f -> do
-        unless (isTestFile f) $ do
-            liftIO $ createDirectoryIfMissing True (takeDirectory (dest </> f))
-            liftIO $ copyFile (cwd </> f) (dest </> f)
-    withReaderT (\nenv -> nenv { testHaveSourceCopy = True }) m
-
--- | If a test needs to modify or write out source files, it's
--- necessary to make a hermetic copy of the source files to operate
--- on.  This function arranges for this to be done in a subdirectory
--- with a given name, so that tests that are sensitive to the path
--- that they're running in (e.g., autoconf tests) can run.
---
--- This requires the test repository to be a Git checkout, because
--- we use the Git metadata to figure out what files to copy into the
--- hermetic copy.
---
--- Also see 'withSourceCopy'.
-withSourceCopyDir :: FilePath -> TestM a -> TestM a
-withSourceCopyDir dir =
-  withReaderT (\nenv -> nenv { testSourceCopyRelativeDir = dir }) . withSourceCopy
 
 -- | Look up the 'InstalledPackageId' of a package name.
 getIPID :: String -> TestM String
@@ -1112,40 +1040,21 @@ withSymlink oldpath newpath0 act = do
 
 writeSourceFile :: FilePath -> String -> TestM ()
 writeSourceFile fp s = do
-    requireHasSourceCopy
     cwd <- fmap testCurrentDir getTestEnv
     liftIO $ writeFile (cwd </> fp) s
 
 copySourceFileTo :: FilePath -> FilePath -> TestM ()
 copySourceFileTo src dest = do
-    requireHasSourceCopy
     cwd <- fmap testCurrentDir getTestEnv
     liftIO $ copyFile (cwd </> src) (cwd </> dest)
 
-requireHasSourceCopy :: TestM ()
-requireHasSourceCopy = do
-    env <- getTestEnv
-    unless (testHaveSourceCopy env) $ do
-        error "This operation requires a source copy; use withSourceCopy and 'git add' all test files"
-
--- NB: Keep this synchronized with partitionTests
-isTestFile :: FilePath -> Bool
-isTestFile f =
-    case takeExtensions f of
-        ".test.hs"      -> True
-        ".multitest.hs" -> True
-        _               -> False
-
 -- | Work around issue #4515 (store paths exceeding the Windows path length
 -- limit) by creating a temporary directory for the new-build store. This
 -- function creates a directory immediately under the current drive on Windows.
 -- The directory must be passed to new- commands with --store-dir.
 withShorterPathForNewBuildStore :: (FilePath -> IO a) -> IO a
-withShorterPathForNewBuildStore test = do
-  tempDir <- if buildOS == Windows
-             then takeDrive `fmap` getCurrentDirectory
-             else getTemporaryDirectory
-  withTempDirectory normal tempDir "cabal-test-store" test
+withShorterPathForNewBuildStore test =
+  withTestDir normal "cabal-test-store" test
 
 -- | Find where a package locates in the store dir. This works only if there is exactly one 1 ghc version
 -- and exactly 1 directory for the given package in the store dir.
@@ -1153,7 +1062,7 @@ findDependencyInStore :: FilePath -- ^store dir
                       -> String -- ^package name prefix
                       -> IO FilePath -- ^package dir
 findDependencyInStore storeDir pkgName = do
-    storeDirForGhcVersion <- head <$> listDirectory storeDir
+    (storeDirForGhcVersion : _) <- listDirectory storeDir
     packageDirs <- listDirectory (storeDir </> storeDirForGhcVersion)
     -- Ideally, we should call 'hashedInstalledPackageId' from 'Distribution.Client.PackageHash'.
     -- But 'PackageHashInputs', especially 'PackageHashConfigInputs', is too hard to construct.
@@ -1162,5 +1071,7 @@ findDependencyInStore storeDir pkgName = do
             then filter (not . flip elem "aeiou") pkgName
                 -- simulates the way 'hashedInstalledPackageId' uses to compress package name
             else pkgName
-    let libDir = head $ filter (pkgName' `isPrefixOf`) packageDirs
+    let libDir = case filter (pkgName' `isPrefixOf`) packageDirs of
+                    [] -> error $ "Could not find " <> pkgName' <> " when searching for " <> pkgName' <> " in\n" <> show packageDirs
+                    (dir:_) -> dir
     pure (storeDir </> storeDirForGhcVersion </> libDir)
