@@ -4,9 +4,14 @@
 # default config
 #######################################################################
 
-HC=ghc-8.2.2
+# We use the default ghc in PATH as default
+# Use the ghc-x.y.z trigger several errors in windows:
+# * It triggers the max path length issue:
+#   See https://github.com/haskell/cabal/issues/6271#issuecomment-1065102255
+# * It triggers a `createProcess: does not exist` error in units tests
+#   See https://github.com/haskell/cabal/issues/8049
+HC=ghc
 CABAL=cabal
-CABALPLAN=cabal-plan
 JOBS=4
 LIBTESTS=true
 CLITESTS=true
@@ -35,8 +40,8 @@ Usage: ./validate.sh [options]
   A script which runs all the tests.
 
 Available options:
-  -j, --jobs JOBS                   cabal v2-build -j argument (default:  $JOBS)
-      --libonly                     Test onlt Cabal-the-library
+  -j, --jobs JOBS                   cabal build -j argument (default:  $JOBS)
+      --libonly                     Test only Cabal-the-library
       --cli                         Test both Cabal-the-library and cabal-install
       --(no-)run-lib-tests          Run library tests
       --(no-)run-cli-tests          Run client tests
@@ -44,7 +49,6 @@ Available options:
       --(no-)run-cli-suite          Run cabal-testsuite with client
   -w, --with-compiler HC            With compiler
       --with-cabal CABAL            With cabal-install
-      --with-cabal-plan CABALPLAN   With cabal-plan
       --extra-hc HC                 Extra compiler to run test-suite with
       --(no-)doctest                Run doctest on library
       --(no-)solver-benchmarks      Build and trial run solver-benchmarks
@@ -52,7 +56,7 @@ Available options:
       --partial-hackage-tests       Run hackage-tests on parts of Hackage data
   -v, --verbose                     Verbose output
   -q, --quiet                       Less output
-  -s, --step STEP                   Run only specific step (can be specified mutliple times)
+  -s, --step STEP                   Run only specific step (can be specified multiple times)
       --list-steps                  List steps and build-targets and exit
       --help                        Print this message and exit
 EOF
@@ -63,17 +67,27 @@ EOF
 
 OUTPUT=$(mktemp)
 
-RED='\033[0;31m'
-GREEN='\033[0;32m'
-BLUE='\033[0;34m'
-CYAN='\033[0;96m'
-RESET='\033[0m' # No Color
+# `red` and `green` are used to output also the spent time in white at the end.
+# `blue` and `cyan` are used to print the spawned command, so they only have one
+# argument.
+red () {
+  printf "\033[0;31m%s\033[0m %s \n" "$1" "$2"
+}
+green () {
+  printf "\033[0;32m%s\033[0m %s \n" "$1" "$2"
+}
+blue () {
+  printf "\033[0;34m%s\033[0m\n" "$1"
+}
+cyan () {
+  printf "\033[0;96m%s\033[0m\n" "$1"
+}
 
 JOB_START_TIME=$(date +%s)
 
 timed() {
     PRETTYCMD=$(echo "$@" | sed -E 's/\/home[^ ]*\/([^\/])/**\/\1/g')
-    echo "$BLUE>>> $PRETTYCMD $RESET"
+    blue "$PRETTYCMD"
     start_time=$(date +%s)
 
     if $VERBOSE; then
@@ -101,7 +115,7 @@ timed() {
             rm -f "$OUTPUT"
         fi
 
-        echo "$GREEN<<< $PRETTYCMD $RESET ($duration/$tduration sec)"
+        green "<<< $PRETTYCMD" "($duration/$tduration sec)"
 
         # bottom-margin
         echo ""
@@ -110,8 +124,8 @@ timed() {
             cat "$OUTPUT"
         fi
 
-        echo "$RED<<< $PRETTYCMD $RESET ($duration/$tduration sec, $RET)"
-        echo "$RED<<< $* $RESET ($duration/$tduration sec, $RET)"
+        red "<<< $PRETTYCMD" "($duration/$tduration sec, $RET)"
+        red "<<< $*" "($duration/$tduration sec, $RET)"
         rm -f "$OUTPUT"
         exit 1
     fi
@@ -120,7 +134,7 @@ timed() {
 print_header() {
     TITLE=$1
     TITLEPAT="$(echo "$TITLE"|sed 's:.:=:g')"
-    echo "$CYAN===X============================================================ $(date +%T) ===$RESET" \
+    cyan "===X========================================================================== $(date +%T) ===" \
       | sed "s#X$TITLEPAT=# $TITLE #"
 
 }
@@ -190,11 +204,6 @@ while [ $# -gt 0 ]; do
             shift
             shift
             ;;
-        --with-cabal-plan)
-            CABALPLAN=$2
-            shift
-            shift
-            ;;
         --extra-hc)
             EXTRAHCS="$EXTRAHCS $2"
             shift
@@ -259,7 +268,6 @@ fi
 
 if [ -z "$STEPS" ]; then
     STEPS="print-config print-tool-versions"
-    if ! $LIBONLY;  then STEPS="$STEPS make-cabal-install-dev"; fi
     STEPS="$STEPS build"
     if $DOCTEST;    then STEPS="$STEPS doctest";   fi
     if $LIBTESTS;   then STEPS="$STEPS lib-tests"; fi
@@ -272,8 +280,8 @@ if [ -z "$STEPS" ]; then
     STEPS="$STEPS time-summary"
 fi
 
-TARGETS="Cabal cabal-testsuite"
-if ! $LIBONLY;  then TARGETS="$TARGETS cabal-install"; fi
+TARGETS="Cabal cabal-testsuite Cabal-tests Cabal-QuickCheck Cabal-tree-diff Cabal-described"
+if ! $LIBONLY;  then TARGETS="$TARGETS cabal-install cabal-install-solver cabal-benchmarks"; fi
 if $BENCHMARKS; then TARGETS="$TARGETS solver-benchmarks"; fi
 
 if $LISTSTEPS; then
@@ -289,13 +297,19 @@ TESTSUITEJOBS="-j$JOBS"
 JOBS="-j$JOBS"
 
 # assume compiler is GHC
-RUNHASKELL=$(echo $HC | sed -E 's/ghc(-[0-9.]*)$/runghc\1/')
+RUNHASKELL=$(echo "$HC" | sed -E 's/ghc(-[0-9.]*)$/runghc\1/')
 
-if [ "$(uname)" = "Linux" ]; then
-    ARCH="x86_64-linux"
-else
-    ARCH="x86_64-osx"
-fi
+case "$(uname)" in
+    MINGW64*)
+        ARCH="x86_64-windows"
+        ;;
+    Linux   )
+        ARCH="x86_64-linux"
+        ;;
+    *)
+        ARCH="x86_64-osx"
+        ;;
+esac
 
 if $LIBONLY; then
     PROJECTFILE=cabal.project.validate.libonly
@@ -307,8 +321,8 @@ BASEHC=ghc-$($HC --numeric-version)
 BUILDDIR=dist-newstyle-validate-$BASEHC
 CABAL_TESTSUITE_BDIR="$(pwd)/$BUILDDIR/build/$ARCH/$BASEHC/cabal-testsuite-3"
 
-CABALNEWBUILD="${CABAL} v2-build $JOBS -w $HC --builddir=$BUILDDIR --project-file=$PROJECTFILE"
-CABALPLANLISTBIN="${CABALPLAN} list-bin --builddir=$BUILDDIR"
+CABALNEWBUILD="${CABAL} build $JOBS -w $HC --builddir=$BUILDDIR --project-file=$PROJECTFILE"
+CABALLISTBIN="${CABAL} list-bin --builddir=$BUILDDIR --project-file=$PROJECTFILE"
 
 # header
 #######################################################################
@@ -318,9 +332,8 @@ print_header print-config
 
 cat <<EOF
 compiler:            $HC
-runhaskell           $RUNHASKELL
+runhaskell:          $RUNHASKELL
 cabal-install:       $CABAL
-cabal-plan:          $CABALPLAN
 jobs:                $JOBS
 Cabal tests:         $LIBTESTS
 cabal-install tests: $CLITESTS
@@ -330,7 +343,7 @@ dependencies only:   $DEPSONLY
 doctest:             $DOCTEST
 benchmarks:          $BENCHMARKS
 verbose:             $VERBOSE
-extra complers:      $EXTRAHCS
+extra compilers:     $EXTRAHCS
 
 EOF
 }
@@ -338,12 +351,11 @@ EOF
 step_print_tool_versions() {
 print_header print-tool-versions
 
-timed $HC --version
-timed $CABAL --version
-timed $CABALPLAN --version
+timed "$HC" --version
+timed "$CABAL" --version
 
 for EXTRAHC in $EXTRAHCS; do
-    timed $EXTRAHC --version
+    timed "$EXTRAHC" --version
 done
 }
 
@@ -353,20 +365,19 @@ step_time_summary() {
     JOB_END_TIME=$(date +%s)
     tduration=$((JOB_END_TIME - JOB_START_TIME))
 
-    echo "$CYAN!!! Validation took $tduration seconds. $RESET"
+    cyan "!!! Validation took $tduration seconds."
 }
 
 # build
 #######################################################################
 
-step_make_cabal_install_dev() {
-print_header "make cabal-install-dev"
-timed ${RUNHASKELL} cabal-dev-scripts/src/Preprocessor.hs -o cabal-install/cabal-install.cabal -f CABAL_FLAG_LIB cabal-install/cabal-install.cabal.pp
-}
-
 step_build() {
 print_header "build"
+print_header "Step Build: dry run"
 timed $CABALNEWBUILD $TARGETS --dry-run || exit 1
+print_header "Step Build: full build plan (cached and to-be-built dependencies):"
+jq -r '."install-plan" | map(."pkg-name" + "-" + ."pkg-version" + " " + ."component-name") | join("\n")' "$BUILDDIR/cache/plan.json"
+print_header "Step Build: actual build"
 timed $CABALNEWBUILD $TARGETS || exit 1
 }
 
@@ -383,24 +394,29 @@ timed doctest -package-env=doctest-Cabal --fast Cabal/Distribution Cabal/Languag
 step_lib_tests() {
 print_header "Cabal: tests"
 
-CMD="$($CABALPLANLISTBIN Cabal:test:unit-tests) $TESTSUITEJOBS --hide-successes --with-ghc=$HC"
-(cd Cabal && timed $CMD) || exit 1
+CMD="$($CABALLISTBIN Cabal-tests:test:unit-tests) $TESTSUITEJOBS --hide-successes --with-ghc=$HC"
+(cd Cabal-tests && timed $CMD) || exit 1
 
-CMD="$($CABALPLANLISTBIN Cabal:test:check-tests) $TESTSUITEJOBS --hide-successes"
-(cd Cabal && timed $CMD) || exit 1
+CMD="$($CABALLISTBIN Cabal-tests:test:check-tests) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
 
-CMD="$($CABALPLANLISTBIN Cabal:test:parser-tests) $TESTSUITEJOBS --hide-successes"
-(cd Cabal && timed $CMD) || exit 1
+CMD="$($CABALLISTBIN Cabal-tests:test:parser-tests) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
 
-CMD=$($CABALPLANLISTBIN Cabal:test:hackage-tests)
-(cd Cabal && timed $CMD read-fields) || exit 1
+CMD="$($CABALLISTBIN Cabal-tests:test:rpmvercmp) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
 
+CMD="$($CABALLISTBIN Cabal-tests:test:no-thunks-test) $TESTSUITEJOBS --hide-successes"
+(cd Cabal-tests && timed $CMD) || exit 1
+
+CMD=$($CABALLISTBIN Cabal-tests:test:hackage-tests)
+(cd Cabal-tests && timed $CMD read-fields) || exit 1
 if $HACKAGETESTSALL; then
-  (cd Cabal && timed $CMD parsec)    || exit 1
-  (cd Cabal && timed $CMD roundtrip) || exit 1
+    (cd Cabal-tests && timed $CMD parsec)    || exit 1
+    (cd Cabal-tests && timed $CMD roundtrip) || exit 1
 else
-  (cd Cabal && timed $CMD parsec d)    || exit 1
-  (cd Cabal && timed $CMD roundtrip k) || exit 1
+    (cd Cabal-tests && timed $CMD parsec d)    || exit 1
+    (cd Cabal-tests && timed $CMD roundtrip k) || exit 1
 fi
 }
 
@@ -410,14 +426,14 @@ fi
 step_lib_suite() {
 print_header "Cabal: cabal-testsuite"
 
-CMD="$($CABALPLANLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR $TESTSUITEJOBS --with-ghc=$HC --hide-successes"
+CMD="$($CABALLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR $TESTSUITEJOBS --with-ghc=$HC --hide-successes"
 (cd cabal-testsuite && timed $CMD) || exit 1
 }
 
 step_lib_suite_extras() {
 for EXTRAHC in $EXTRAHCS; do
 
-CMD="$($CABALPLANLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR $TESTSUITEJOBS --with-ghc=$EXTRAHC --hide-successes"
+CMD="$($CABALLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR $TESTSUITEJOBS --with-ghc=$EXTRAHC --hide-successes"
 (cd cabal-testsuite && timed $CMD) || exit 1
 
 done
@@ -430,19 +446,19 @@ step_cli_tests() {
 print_header "cabal-install: tests"
 
 # this are sorted in asc time used, quicker tests first.
-CMD="$($CABALPLANLISTBIN cabal-install:test:solver-quickcheck) $TESTSUITEJOBS --hide-successes"
+CMD="$($CABALLISTBIN cabal-install:test:long-tests) $TESTSUITEJOBS --hide-successes"
 (cd cabal-install && timed $CMD) || exit 1
 
 # This doesn't work in parallel either
-CMD="$($CABALPLANLISTBIN cabal-install:test:unit-tests) -j1 --hide-successes"
+CMD="$($CABALLISTBIN cabal-install:test:unit-tests) -j1 --hide-successes"
 (cd cabal-install && timed $CMD) || exit 1
 
 # Only single job, otherwise we fail with "Heap exhausted"
-CMD="$($CABALPLANLISTBIN cabal-install:test:memory-usage-tests) -j1 --hide-successes"
+CMD="$($CABALLISTBIN cabal-install:test:mem-use-tests) -j1 --hide-successes"
 (cd cabal-install && timed $CMD) || exit 1
 
 # This test-suite doesn't like concurrency
-CMD="$($CABALPLANLISTBIN cabal-install:test:integration-tests2) -j1 --hide-successes --with-ghc=$HC"
+CMD="$($CABALLISTBIN cabal-install:test:integration-tests2) -j1 --hide-successes --with-ghc=$HC"
 (cd cabal-install && timed $CMD) || exit 1
 }
 
@@ -452,7 +468,7 @@ CMD="$($CABALPLANLISTBIN cabal-install:test:integration-tests2) -j1 --hide-succe
 step_cli_suite() {
 print_header "cabal-install: cabal-testsuite"
 
-CMD="$($CABALPLANLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR --with-cabal=$($CABALPLANLISTBIN cabal-install:exe:cabal) $TESTSUITEJOBS --hide-successes"
+CMD="$($CABALLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TESTSUITE_BDIR --with-cabal=$($CABALLISTBIN cabal-install:exe:cabal) $TESTSUITEJOBS  --with-ghc=$HC --hide-successes"
 (cd cabal-testsuite && timed $CMD) || exit 1
 }
 
@@ -462,7 +478,7 @@ CMD="$($CABALPLANLISTBIN cabal-testsuite:exe:cabal-tests) --builddir=$CABAL_TEST
 step_solver_benchmarks_tests() {
 print_header "solver-benchmarks: test"
 
-CMD="$($CABALPLANLISTBIN solver-benchmarks:test:unit-tests)"
+CMD="$($CABALLISTBIN solver-benchmarks:test:unit-tests)"
 (cd Cabal && timed $CMD) || exit 1
 }
 
@@ -470,7 +486,7 @@ step_solver_benchmarks_run() {
 print_header "solver-benchmarks: run"
 
 SOLVEPKG=Chart-diagrams
-CMD="$($CABALPLANLISTBIN solver-benchmarks:exe:hackage-benchmark) --cabal1=$CABAL --cabal2=$($CABALPLANLISTBIN cabal-install:exe:cabal) --trials=5 --packages=$SOLVEPKG --print-trials"
+CMD="$($CABALLISTBIN solver-benchmarks:exe:hackage-benchmark) --cabal1=$CABAL --cabal2=$($CABALLISTBIN cabal-install:exe:cabal) --trials=5 --packages=$SOLVEPKG --print-trials"
 (cd Cabal && timed $CMD) || exit 1
 }
 
@@ -481,7 +497,6 @@ for step in $STEPS; do
     case $step in
         print-config)             step_print_config            ;;
         print-tool-versions)      step_print_tool_versions     ;;
-        make-cabal-install-dev)   step_make_cabal_install_dev  ;;
         build)                    step_build                   ;;
         doctest)                  step_doctest                 ;;
         lib-tests)                step_lib_tests               ;;
@@ -492,7 +507,7 @@ for step in $STEPS; do
         solver-benchmarks-tests)  step_solver_benchmarks_tests ;;
         solver-benchmarks-run)    step_solver_benchmarks_run   ;;
         time-summary)             step_time_summary            ;;
-        *) 
+        *)
             echo "Invalid step $step"
             exit 1
             ;;
