@@ -13,6 +13,7 @@
 
 module Tests.Properties.LowLevel (testLowLevel) where
 
+import Prelude hiding (head, tail)
 import Control.Applicative ((<$>), pure)
 import Control.Exception as E (SomeException, catch, evaluate)
 import Data.Int (Int32, Int64)
@@ -30,6 +31,7 @@ import qualified Data.Text as T
 import qualified Data.Text.IO as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.IO as TL
+import qualified Data.Text.IO.Utf8 as TU
 import qualified System.IO as IO
 
 #ifdef MIN_VERSION_tasty_inspection_testing
@@ -101,11 +103,14 @@ t_literal_foo = T.pack "foo"
 t_write_read = write_read T.unlines T.filter T.hPutStr T.hGetContents
 tl_write_read = write_read TL.unlines TL.filter TL.hPutStr TL.hGetContents
 
-t_write_read_line m b t = write_read head T.filter T.hPutStrLn
+t_write_read_line m b t = write_read (T.concat . take 1) T.filter T.hPutStrLn
                             T.hGetLine m b [t]
-tl_write_read_line m b t = write_read head TL.filter TL.hPutStrLn
+tl_write_read_line m b t = write_read (TL.concat . take 1) TL.filter TL.hPutStrLn
                              TL.hGetLine m b [t]
 
+utf8_write_read = write_read T.unlines T.filter TU.hPutStr TU.hGetContents
+utf8_write_read_line m b t = write_read (T.concat . take 1) T.filter TU.hPutStrLn
+                            TU.hGetLine m b [t]
 
 testLowLevel :: TestTree
 testLowLevel =
@@ -141,7 +146,9 @@ testLowLevel =
       testProperty "t_write_read" t_write_read,
       testProperty "tl_write_read" tl_write_read,
       testProperty "t_write_read_line" t_write_read_line,
-      testProperty "tl_write_read_line" tl_write_read_line
+      testProperty "tl_write_read_line" tl_write_read_line,
+      testProperty "utf8_write_read" utf8_write_read,
+      testProperty "utf8_write_read_line" utf8_write_read_line
       -- These tests are subject to I/O race conditions
       -- testProperty "t_put_get" t_put_get,
       -- testProperty "tl_put_get" tl_put_get
