@@ -165,6 +165,18 @@ ObjectCode *loaded_objects;
 // map static closures to their ObjectCode.
 static OCSectionIndices *global_s_indices = NULL;
 
+// Is it safe for us to unload code?
+static bool safeToUnload(void)
+{
+    if (RtsFlags.ProfFlags.doHeapProfile != NO_HEAP_PROFILING) {
+        // We mustn't unload anything as the heap census may contain
+        // references into static data (e.g. cost centre names).
+        // See #24512.
+        return false;
+    }
+    return true;
+}
+
 static OCSectionIndices *createOCSectionIndices(void)
 {
     // TODO (osa): Maybe initialize as empty (without allocation) and allocate
@@ -186,12 +198,12 @@ static void freeOCSectionIndices(OCSectionIndices *s_indices)
     stgFree(s_indices);
 }
 
-void initUnloadCheck()
+void initUnloadCheck(void)
 {
     global_s_indices = createOCSectionIndices();
 }
 
-void exitUnloadCheck()
+void exitUnloadCheck(void)
 {
     freeOCSectionIndices(global_s_indices);
     global_s_indices = NULL;
@@ -436,7 +448,7 @@ void markObjectCode(const void *addr)
 
 // Returns whether or not the GC that follows needs to mark code for potential
 // unloading.
-bool prepareUnloadCheck()
+bool prepareUnloadCheck(void)
 {
     if (global_s_indices == NULL) {
         return false;
@@ -453,10 +465,12 @@ bool prepareUnloadCheck()
     return true;
 }
 
-void checkUnload()
+void checkUnload(void)
 {
     if (global_s_indices == NULL) {
         return;
+    } else if (!safeToUnload()) {
+        return;
     }
 
     // At this point we've marked all dynamically loaded static objects
@@ -478,8 +492,6 @@ void checkUnload()
         next = oc->next;
         ASSERT(oc->status == OBJECT_UNLOADED);
 
-        removeOCSectionIndices(s_indices, oc);
-
         // Symbols should be removed by unloadObj_.
         // NB (osa): If this assertion doesn't hold then freeObjectCode below
         // will corrupt symhash as keys of that table live in ObjectCodes. If
@@ -487,8 +499,17 @@ void checkUnload()
         // RTS) then it's probably because this assertion did not hold.
         ASSERT(oc->symbols == NULL);
 
-        freeObjectCode(oc);
-        n_unloaded_objects -= 1;
+        if (oc->unloadable) {
+            removeOCSectionIndices(s_indices, oc);
+            freeObjectCode(oc);
+            n_unloaded_objects -= 1;
+        } else {
+            // If we don't have enough information to
+            // accurately determine the reachability of
+            // the object then hold onto it.
+            oc->next = objects;
+            objects = oc;
+        }
     }
 
     old_objects = NULL;
