@@ -8,7 +8,7 @@
  * pp. 446 -- 457.
  * -------------------------------------------------------------------------- */
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 
 #include "Hash.h"
@@ -18,13 +18,15 @@
    since we compile these things these days with cabal we can no longer
    specify optimization per file.  So we have to resort to pragmas.  */
 #if defined(__GNUC__) || defined(__GNUG__)
+#if !defined(__clang__)
+#if !defined(DEBUG)
 #pragma GCC push_options
 #pragma GCC optimize ("O3")
 #endif
+#endif
+#endif
 
-#define XXH_NAMESPACE __rts_
-#define XXH_STATIC_LINKING_ONLY   /* access advanced declarations */
-#define XXH_PRIVATE_API
+#define XXH_INLINE_ALL
 
 #include "xxhash.h"
 
@@ -48,7 +50,6 @@ typedef struct hashlist {
 } HashList;
 
 typedef struct chunklist {
-  HashList *chunk;
   struct chunklist *next;
 } HashListChunk;
 
@@ -61,23 +62,26 @@ struct hashtable {
     int bcount;             /* Number of buckets */
     HashList **dir[HDIRSIZE];   /* Directory of segments */
     HashList *freeList;         /* free list of HashLists */
-    HashListChunk *chunks;
-    HashFunction *hash;         /* hash function */
-    CompareFunction *compare;   /* key comparison function */
+    HashListChunk *chunks;      /* list of HashListChunks so we can later free them */
 };
 
+/* Create an identical structure, but is distinct on a type level,
+ * for string hash table. Since it's a direct embedding of
+ * a hashtable and not a reference, there shouldn't be
+ * any overhead post-compilation.  */
+struct strhashtable { struct hashtable table; };
+
 /* -----------------------------------------------------------------------------
  * Hash first using the smaller table.  If the bucket is less than the
  * next bucket to be split, re-hash using the larger table.
  * -------------------------------------------------------------------------- */
-
 int
 hashWord(const HashTable *table, StgWord key)
 {
     int bucket;
 
     /* Strip the boring zero bits */
-    key /= sizeof(StgWord);
+    key >>= sizeof(StgWord);
 
     /* Mod the size of the hash table (a power of 2) */
     bucket = key & table->mask1;
@@ -93,8 +97,8 @@ int
 hashStr(const HashTable *table, StgWord w)
 {
     const char *key = (char*) w;
-#if defined(x86_64_HOST_ARCH)
-    StgWord h = XXH64 (key, strlen(key), 1048583);
+#if WORD_SIZE_IN_BITS == 64
+    StgWord h = XXH3_64bits_withSeed (key, strlen(key), 1048583);
 #else
     StgWord h = XXH32 (key, strlen(key), 1048583);
 #endif
@@ -110,13 +114,13 @@ hashStr(const HashTable *table, StgWord w)
     return bucket;
 }
 
-static int
+STATIC_INLINE int
 compareWord(StgWord key1, StgWord key2)
 {
     return (key1 == key2);
 }
 
-static int
+STATIC_INLINE int
 compareStr(StgWord key1, StgWord key2)
 {
     return (strcmp((char *)key1, (char *)key2) == 0);
@@ -127,7 +131,7 @@ compareStr(StgWord key1, StgWord key2)
  * Allocate a new segment of the dynamically growing hash table.
  * -------------------------------------------------------------------------- */
 
-static void
+STATIC_INLINE void
 allocSegment(HashTable *table, int segment)
 {
     table->dir[segment] = stgMallocBytes(HSEGSIZE * sizeof(HashList *),
@@ -141,8 +145,8 @@ allocSegment(HashTable *table, int segment)
  * by @table->split@ is affected by the expansion.
  * -------------------------------------------------------------------------- */
 
-static void
-expand(HashTable *table)
+STATIC_INLINE void
+expand(HashTable *table, HashFunction f)
 {
     int oldsegment;
     int oldindex;
@@ -183,7 +187,7 @@ expand(HashTable *table)
     old = new = NULL;
     for (hl = table->dir[oldsegment][oldindex]; hl != NULL; hl = next) {
         next = hl->next;
-        if (table->hash(table, hl->key) == newbucket) {
+        if (f(table, hl->key) == newbucket) {
             hl->next = new;
             new = hl;
         } else {
@@ -197,19 +201,20 @@ expand(HashTable *table)
     return;
 }
 
-void *
-lookupHashTable(const HashTable *table, StgWord key)
+STATIC_INLINE void*
+lookupHashTable_inlined(const HashTable *table, StgWord key,
+                        HashFunction f, CompareFunction cmp)
 {
     int bucket;
     int segment;
     int index;
+
     HashList *hl;
 
-    bucket = table->hash(table, key);
+    bucket = f(table, key);
     segment = bucket / HSEGSIZE;
     index = bucket % HSEGSIZE;
 
-    CompareFunction *cmp = table->compare;
     for (hl = table->dir[segment][index]; hl != NULL; hl = hl->next) {
         if (cmp(hl->key, key))
             return (void *) hl->data;
@@ -219,6 +224,26 @@ lookupHashTable(const HashTable *table, StgWord key)
     return NULL;
 }
 
+void *
+lookupHashTable_(const HashTable *table, StgWord key,
+                 HashFunction f, CompareFunction cmp)
+{
+    return lookupHashTable_inlined(table, key, f, cmp);
+}
+
+void *
+lookupHashTable(const HashTable *table, StgWord key)
+{
+    return lookupHashTable_inlined(table, key, hashWord, compareWord);
+}
+
+void *
+lookupStrHashTable(const StrHashTable* table, const char* key)
+{
+    return lookupHashTable_inlined(&table->table, (StgWord) key,
+                                   hashStr, compareStr);
+}
+
 // Puts up to szKeys keys of the hash table into the given array. Returns the
 // actual amount of keys that have been retrieved.
 //
@@ -253,41 +278,56 @@ int keysHashTable(HashTable *table, StgWord keys[], int szKeys) {
 /* -----------------------------------------------------------------------------
  * We allocate the hashlist cells in large chunks to cut down on malloc
  * overhead.  Although we keep a free list of hashlist cells, we make
- * no effort to actually return the space to the malloc arena.
+ * no effort to actually return the space to the malloc arena. Eventually
+ * they will all be freed when we free the HashListChunks.
  * -------------------------------------------------------------------------- */
 
 static HashList *
 allocHashList (HashTable *table)
 {
-    HashList *hl, *p;
-    HashListChunk *cl;
-
-    if ((hl = table->freeList) != NULL) {
+    if (table->freeList != NULL) {
+        HashList *hl = table->freeList;
         table->freeList = hl->next;
+        return hl;
     } else {
-        hl = stgMallocBytes(HCHUNK * sizeof(HashList), "allocHashList");
-        cl = stgMallocBytes(sizeof (*cl), "allocHashList: chunkList");
-        cl->chunk = hl;
+        /* We allocate one block of memory which contains:
+         *
+         *  1. A HashListChunk, which gets linked onto HashTable.chunks.
+         *     This forms a list of all chunks associated with the HashTable
+         *     and is what we will free when we free the HashTable.
+         *
+         *  2. Several HashLists. One of these will get returned. The rest are
+         *     placed on the freeList.
+         *
+         */
+        HashListChunk *cl = stgMallocBytes(sizeof(HashListChunk) + HCHUNK * sizeof(HashList), "allocHashList");
+        HashList *hl = (HashList *) &cl[1];
         cl->next = table->chunks;
         table->chunks = cl;
 
         table->freeList = hl + 1;
-        for (p = table->freeList; p < hl + HCHUNK - 1; p++)
+        HashList *p = table->freeList;
+        for (; p < hl + HCHUNK - 1; p++)
             p->next = p + 1;
         p->next = NULL;
+        return hl;
     }
-    return hl;
 }
 
 static void
 freeHashList (HashTable *table, HashList *hl)
 {
+    // We place the HashList on the freeList. We make no attempt to bound the
+    // size of the free list for the time being. The HashLists on the freeList
+    // are freed when the HashTable itself is freed as a result of freeing the
+    // HashListChunks.
     hl->next = table->freeList;
     table->freeList = hl;
 }
 
-void
-insertHashTable(HashTable *table, StgWord key, const void *data)
+STATIC_INLINE void
+insertHashTable_inlined(HashTable *table, StgWord key,
+                        const void *data, HashFunction f)
 {
     int bucket;
     int segment;
@@ -300,9 +340,9 @@ insertHashTable(HashTable *table, StgWord key, const void *data)
 
     /* When the average load gets too high, we expand the table */
     if (++table->kcount >= HLOAD * table->bcount)
-        expand(table);
+        expand(table, f);
 
-    bucket = table->hash(table, key);
+    bucket = f(table, key);
     segment = bucket / HSEGSIZE;
     index = bucket % HSEGSIZE;
 
@@ -312,11 +352,30 @@ insertHashTable(HashTable *table, StgWord key, const void *data)
     hl->data = data;
     hl->next = table->dir[segment][index];
     table->dir[segment][index] = hl;
-
 }
 
-void *
-removeHashTable(HashTable *table, StgWord key, const void *data)
+void
+insertHashTable_(HashTable *table, StgWord key,
+                 const void *data, HashFunction f)
+{
+    return insertHashTable_inlined(table, key, data, f);
+}
+
+void
+insertHashTable(HashTable *table, StgWord key, const void *data)
+{
+    insertHashTable_inlined(table, key, data, hashWord);
+}
+
+void
+insertStrHashTable(StrHashTable *table, const char * key, const void *data)
+{
+    insertHashTable_inlined(&table->table, (StgWord) key, data, hashStr);
+}
+
+STATIC_INLINE void*
+removeHashTable_inlined(HashTable *table, StgWord key, const void *data,
+                        HashFunction f, CompareFunction cmp)
 {
     int bucket;
     int segment;
@@ -324,12 +383,12 @@ removeHashTable(HashTable *table, StgWord key, const void *data)
     HashList *hl;
     HashList *prev = NULL;
 
-    bucket = table->hash(table, key);
+    bucket = f(table, key);
     segment = bucket / HSEGSIZE;
     index = bucket % HSEGSIZE;
 
     for (hl = table->dir[segment][index]; hl != NULL; hl = hl->next) {
-        if (table->compare(hl->key,key) && (data == NULL || hl->data == data)) {
+        if (cmp(hl->key, key) && (data == NULL || hl->data == data)) {
             if (prev == NULL)
                 table->dir[segment][index] = hl->next;
             else
@@ -346,6 +405,26 @@ removeHashTable(HashTable *table, StgWord key, const void *data)
     return NULL;
 }
 
+void*
+removeHashTable_(HashTable *table, StgWord key, const void *data,
+                 HashFunction f, CompareFunction cmp)
+{
+    return removeHashTable_inlined(table, key, data, f, cmp);
+}
+
+void *
+removeHashTable(HashTable *table, StgWord key, const void *data)
+{
+    return removeHashTable_inlined(table, key, data, hashWord, compareWord);
+}
+
+void *
+removeStrHashTable(StrHashTable *table, const char * key, const void *data)
+{
+    return removeHashTable_inlined(&table->table, (StgWord) key,
+                                   data, hashStr, compareStr);
+}
+
 /* -----------------------------------------------------------------------------
  * When we free a hash table, we are also good enough to free the
  * data part of each (key, data) pair, as long as our caller can tell
@@ -355,34 +434,35 @@ removeHashTable(HashTable *table, StgWord key, const void *data)
 void
 freeHashTable(HashTable *table, void (*freeDataFun)(void *) )
 {
-    long segment;
-    long index;
-    HashList *hl;
-    HashList *next;
-    HashListChunk *cl, *cl_next;
-
     /* The last bucket with something in it is table->max + table->split - 1 */
-    segment = (table->max + table->split - 1) / HSEGSIZE;
-    index = (table->max + table->split - 1) % HSEGSIZE;
+    long segment = (table->max + table->split - 1) / HSEGSIZE;
+    long index = (table->max + table->split - 1) % HSEGSIZE;
 
+    /* Free table segments */
     while (segment >= 0) {
-        while (index >= 0) {
-            for (hl = table->dir[segment][index]; hl != NULL; hl = next) {
-                next = hl->next;
-                if (freeDataFun != NULL)
+        if (freeDataFun) {
+            while (index >= 0) {
+                HashList *next;
+                for (HashList *hl = table->dir[segment][index]; hl != NULL; hl = next) {
+                    next = hl->next;
                     (*freeDataFun)((void *) hl->data);
+                }
+                index--;
             }
-            index--;
         }
         stgFree(table->dir[segment]);
         segment--;
         index = HSEGSIZE - 1;
     }
-    for (cl = table->chunks; cl != NULL; cl = cl_next) {
-        cl_next = cl->next;
-        stgFree(cl->chunk);
-        stgFree(cl);
+
+    /* Free chunks */
+    HashListChunk *cl = table->chunks;
+    while (cl != NULL) {
+        HashListChunk *old = cl;
+        cl = cl->next;
+        stgFree(old);
     }
+
     stgFree(table);
 }
 
@@ -393,17 +473,13 @@ freeHashTable(HashTable *table, void (*freeDataFun)(void *) )
 void
 mapHashTable(HashTable *table, void *data, MapHashFn fn)
 {
-    long segment;
-    long index;
-    HashList *hl;
-
     /* The last bucket with something in it is table->max + table->split - 1 */
-    segment = (table->max + table->split - 1) / HSEGSIZE;
-    index = (table->max + table->split - 1) % HSEGSIZE;
+    long segment = (table->max + table->split - 1) / HSEGSIZE;
+    long index = (table->max + table->split - 1) % HSEGSIZE;
 
     while (segment >= 0) {
         while (index >= 0) {
-            for (hl = table->dir[segment][index]; hl != NULL; hl = hl->next) {
+            for (HashList *hl = table->dir[segment][index]; hl != NULL; hl = hl->next) {
                 fn(data, hl->key, hl->data);
             }
             index--;
@@ -413,6 +489,25 @@ mapHashTable(HashTable *table, void *data, MapHashFn fn)
     }
 }
 
+void
+mapHashTableKeys(HashTable *table, void *data, MapHashFnKeys fn)
+{
+    /* The last bucket with something in it is table->max + table->split - 1 */
+    long segment = (table->max + table->split - 1) / HSEGSIZE;
+    long index = (table->max + table->split - 1) % HSEGSIZE;
+
+    while (segment >= 0) {
+        while (index >= 0) {
+            for (HashList *hl = table->dir[segment][index]; hl != NULL; hl = hl->next) {
+                fn(data, &hl->key, hl->data);
+            }
+            index--;
+        }
+        segment--;
+        index = HSEGSIZE - 1;
+    }
+}
+
 void
 iterHashTable(HashTable *table, void *data, IterHashFn fn)
 {
@@ -440,7 +535,7 @@ iterHashTable(HashTable *table, void *data, IterHashFn fn)
  * -------------------------------------------------------------------------- */
 
 HashTable *
-allocHashTable_(HashFunction *hash, CompareFunction *compare)
+allocHashTable(void)
 {
     HashTable *table;
     HashList **hb;
@@ -460,25 +555,20 @@ allocHashTable_(HashFunction *hash, CompareFunction *compare)
     table->bcount = HSEGSIZE;
     table->freeList = NULL;
     table->chunks = NULL;
-    table->hash = hash;
-    table->compare = compare;
 
     return table;
 }
 
-HashTable *
-allocHashTable(void)
-{
-    return allocHashTable_(hashWord, compareWord);
-}
-
-HashTable *
-allocStrHashTable(void)
-{
-    return allocHashTable_(hashStr, compareStr);
-}
-
 int keyCountHashTable (HashTable *table)
 {
     return table->kcount;
 }
+
+
+#if defined(__GNUC__) || defined(__GNUG__)
+#if !defined(__clang__)
+#if !defined(DEBUG)
+#pragma GCC pop_options
+#endif
+#endif
+#endif
