@@ -4,34 +4,11 @@
  * Copyright (c) The GHC Team, 1994-2002.
  * ---------------------------------------------------------------------------*/
 
-/*
-Note [CBV Functions and the interpreter]
-~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-When the byte code interpreter loads a reference to a value it often
-ends up as a non-tagged pointers *especially* if we already know a value
-is a certain constructor and therefore don't perform an eval on the reference.
-This causes friction with CBV functions which assume
-their value arguments are properly tagged by the caller.
-
-In order to ensure CBV functions still get passed tagged functions we have
-three options:
-a)  Special case the interpreter behaviour into the tag inference analysis.
-    If we assume the interpreter can't properly tag value references the STG passes
-    would then wrap such calls in appropriate evals which are executed at runtime.
-    This would ensure tags by doing additional evals at runtime.
-b)  When the interpreter pushes references for known constructors instead of
-    pushing the objects address add the tag to the value pushed. This is what
-    the NCG backends do.
-c)  When the interpreter pushes a reference inspect the closure of the object
-    and apply the appropriate tag at runtime.
-
-For now we use approach c). Mostly because it's easiest to implement. We also don't
-tag functions as tag inference currently doesn't rely on those being properly tagged.
-*/
 
 #include "rts/PosixSource.h"
 #include "Rts.h"
 #include "RtsAPI.h"
+#include "RtsFlags.h"
 #include "rts/Bytecodes.h"
 
 // internal headers
@@ -325,7 +302,7 @@ static StgWord app_ptrs_itbl[] = {
 };
 
 HsStablePtr rts_breakpoint_io_action; // points to the IO action which is executed on a breakpoint
-                                      // it is set in compiler/GHC.hs:runStmt
+                                      // it is set in ghci/GHCi/Run.hs:withBreakAction
 
 Capability *
 interpretBCO (Capability* cap)
@@ -339,8 +316,9 @@ interpretBCO (Capability* cap)
 
     LOAD_THREAD_STATE();
 
-    cap->r.rHpLim = (P_)1; // HpLim is the context-switch flag; when it
-                           // goes to zero we must return to the scheduler.
+    // N.B. HpLim is the context-switch flag; when it
+    // goes to zero we must return to the scheduler.
+    RELAXED_STORE_ALWAYS(&cap->r.rHpLim, (P_)1);
 
     IF_DEBUG(interpreter,
              debugBelch(
@@ -1111,7 +1089,7 @@ run_BCO:
         /* check for a breakpoint on the beginning of a let binding */
         case bci_BRK_FUN:
         {
-            int arg1_brk_array, arg2_array_index, arg3_module_uniq;
+            int arg1_brk_array, arg2_array_index, arg3_module_name;
 #if defined(PROFILING)
             int arg4_cc;
 #endif
@@ -1129,7 +1107,7 @@ run_BCO:
 
             arg1_brk_array      = BCO_GET_LARGE_ARG;
             arg2_array_index    = BCO_NEXT;
-            arg3_module_uniq    = BCO_GET_LARGE_ARG;
+            arg3_module_name    = BCO_GET_LARGE_ARG;
 #if defined(PROFILING)
             arg4_cc             = BCO_GET_LARGE_ARG;
 #else
@@ -1194,7 +1172,7 @@ run_BCO:
                   // continue execution of this BCO when the IO action returns.
                   //
                   // ioAction :: Int#        -- the breakpoint index
-                  //          -> Int#        -- the module uniq
+                  //          -> Addr#       -- the breakpoint module
                   //          -> Bool        -- exception?
                   //          -> HValue      -- the AP_STACK, or exception
                   //          -> IO ()
@@ -1208,7 +1186,7 @@ run_BCO:
                   SpW(8)  = (W_)new_aps;
                   SpW(7)  = (W_)False_closure;         // True <=> an exception
                   SpW(6)  = (W_)&stg_ap_ppv_info;
-                  SpW(5)  = (W_)BCO_LIT(arg3_module_uniq);
+                  SpW(5)  = (W_)BCO_LIT(arg3_module_name);
                   SpW(4)  = (W_)&stg_ap_n_info;
                   SpW(3)  = (W_)arg2_array_index;
                   SpW(2)  = (W_)&stg_ap_n_info;
@@ -1299,63 +1277,28 @@ run_BCO:
 
         case bci_PUSH8_W: {
             W_ off = BCO_GET_LARGE_ARG;
-            *(StgWord*)(Sp_minusW(1)) = *(StgWord8*)(Sp_plusB(off));
+            *(StgWord8*)(Sp_minusW(1)) = *(StgWord8*)(Sp_plusB(off));
             Sp_subW(1);
             goto nextInsn;
         }
 
         case bci_PUSH16_W: {
             W_ off = BCO_GET_LARGE_ARG;
-            *(StgWord*)(Sp_minusW(1)) = *(StgWord16*)(Sp_plusB(off));
+            *(StgWord16*)(Sp_minusW(1)) = *(StgWord16*)(Sp_plusB(off));
             Sp_subW(1);
             goto nextInsn;
         }
 
         case bci_PUSH32_W: {
             W_ off = BCO_GET_LARGE_ARG;
-            *(StgWord*)(Sp_minusW(1)) = *(StgWord32*)(Sp_plusB(off));
+            *(StgWord32*)(Sp_minusW(1)) = *(StgWord32*)(Sp_plusB(off));
             Sp_subW(1);
             goto nextInsn;
         }
 
         case bci_PUSH_G: {
             W_ o1 = BCO_GET_LARGE_ARG;
-            StgClosure *tagged_obj = (StgClosure*) BCO_PTR(o1);
-
-            tag_push_g:
-            ASSERT(LOOKS_LIKE_CLOSURE_PTR((StgClosure*) tagged_obj));
-            // Here we make sure references we push are tagged.
-            // See Note [CBV Functions and the interpreter] in Info.hs
-
-            //Safe some memory reads if we already have a tag.
-            if(GET_CLOSURE_TAG(tagged_obj) == 0) {
-                StgClosure *obj = UNTAG_CLOSURE(tagged_obj);
-                switch ( get_itbl(obj)->type ) {
-                    case IND:
-                    case IND_STATIC:
-                    {
-                        tagged_obj = ACQUIRE_LOAD(&((StgInd*)obj)->indirectee);
-                        goto tag_push_g;
-                    }
-                    case CONSTR:
-                    case CONSTR_1_0:
-                    case CONSTR_0_1:
-                    case CONSTR_2_0:
-                    case CONSTR_1_1:
-                    case CONSTR_0_2:
-                    case CONSTR_NOCAF:
-                        // The value is already evaluated, so we can just return it. However,
-                        // before we do, we MUST ensure that the pointer is tagged, because we
-                        // might return to a native `case` expression, which assumes the returned
-                        // pointer is tagged so it can use the tag to select an alternative.
-                        tagged_obj = tagConstr(obj);
-                        break;
-                    default:
-                        break;
-                }
-            }
-
-            SpW(-1) = (W_) tagged_obj;
+            SpW(-1) = BCO_PTR(o1);
             Sp_subW(1);
             goto nextInsn;
         }
@@ -2026,7 +1969,7 @@ run_BCO:
             // context switching: sometimes the scheduler can invoke
             // the interpreter with context_switch == 1, particularly
             // if the -C0 flag has been given on the cmd line.
-            if (cap->r.rHpLim == NULL) {
+            if (RELAXED_LOAD(&cap->r.rHpLim) == NULL) {
                 Sp_subW(1); SpW(0) = (W_)&stg_enter_info;
                 RETURN_TO_SCHEDULER(ThreadInterpret, ThreadYielding);
             }
