@@ -142,6 +142,23 @@ restore_locale( void )
 unsigned int era;
 static uint32_t max_era;
 
+StgWord user_era;
+
+void
+setUserEra (StgWord w){
+  user_era = w;
+}
+
+StgWord
+getUserEra (void){
+  return user_era;
+}
+
+StgWord
+incrementUserEra (StgWord w){
+  return atomic_inc(&user_era, w);
+}
+
 inline void
 initLDVCtr( counter *ctr )
 {
@@ -180,6 +197,11 @@ closureIdentity( const StgClosure *p )
         return p->header.prof.ccs->cc->module;
     case HEAP_BY_DESCR:
         return GET_PROF_DESC(get_itbl(p));
+    case HEAP_BY_ERA:
+        // Static objects should have user_era = 0
+        // MP: If user_era == 0 then closureIdentity returns the NULL pointer, and
+        // the closure is not counted to the census
+        return (void *)p->header.prof.hp.era;
     case HEAP_BY_TYPE:
         return GET_PROF_TYPE(get_itbl(p));
     case HEAP_BY_RETAINER:
@@ -220,21 +242,6 @@ closureIdentity( const StgClosure *p )
 /* --------------------------------------------------------------------------
  * Profiling type predicates
  * ----------------------------------------------------------------------- */
-#if defined(PROFILING)
-STATIC_INLINE bool
-doingLDVProfiling( void )
-{
-    return (RtsFlags.ProfFlags.doHeapProfile == HEAP_BY_LDV
-            || RtsFlags.ProfFlags.bioSelector != NULL);
-}
-
-bool
-doingRetainerProfiling( void )
-{
-    return (RtsFlags.ProfFlags.doHeapProfile == HEAP_BY_RETAINER
-            || RtsFlags.ProfFlags.retainerSelector != NULL);
-}
-#endif /* PROFILING */
 
 // Processes a closure 'c' being destroyed whose size is 'size'.
 // Make sure that LDV_recordDead() is not invoked on 'inherently used' closures
@@ -356,6 +363,10 @@ freeEra(Census *census)
 static void
 nextEra( void )
 {
+    if (user_era > 0 && RtsFlags.ProfFlags.incrementUserEra){
+      user_era++;
+    }
+
 #if defined(PROFILING)
     if (doingLDVProfiling()) {
         era++;
@@ -419,7 +430,7 @@ void freeHeapProfiling (void)
 }
 
 /* --------------------------------------------------------------------------
- * Initialize the heap profilier
+ * Initialize the heap profiler
  * ----------------------------------------------------------------------- */
 void
 initHeapProfiling(void)
@@ -437,18 +448,14 @@ initHeapProfiling(void)
         stem = stgMallocBytes(strlen(RtsFlags.CcFlags.outputFileNameStem) + 1, "initHeapProfiling");
         strcpy(stem, RtsFlags.CcFlags.outputFileNameStem);
     } else {
-
         stem = stgMallocBytes(strlen(prog_name) + 1, "initHeapProfiling");
         strcpy(stem, prog_name);
+
+        // Drop the platform's executable suffix if there is one
 #if defined(mingw32_HOST_OS)
-            // on Windows, drop the .exe suffix if there is one
-            {
-                char *suff;
-                suff = strrchr(stem,'.');
-                if (suff != NULL && !strcmp(suff,".exe")) {
-                    *suff = '\0';
-                }
-            }
+        dropExtension(stem, ".exe");
+#elif defined(wasm32_HOST_ARCH)
+        dropExtension(stem, ".wasm");
 #endif
     }
 
@@ -483,6 +490,14 @@ initHeapProfiling(void)
 #endif
 #endif
 
+#if defined(PROFILING)
+    if (doingErasProfiling()){
+      user_era = 1;
+    }
+#else
+    user_era = 0;
+#endif
+
     // we only count eras if we're doing LDV profiling.  Otherwise era
     // is fixed at zero.
 #if defined(PROFILING)
@@ -703,6 +718,9 @@ closureSatisfiesConstraints( const StgClosure* p )
                                 RtsFlags.ProfFlags.typeSelector );
        if (!b) return false;
    }
+   if (RtsFlags.ProfFlags.eraSelector) {
+      return (p->header.prof.hp.era == RtsFlags.ProfFlags.eraSelector);
+   }
    if (RtsFlags.ProfFlags.retainerSelector) {
        RetainerSet *rs;
        uint32_t i;
@@ -939,6 +957,12 @@ dumpCensus( Census *census )
             traceHeapProfSampleCostCentre(0, (CostCentreStack *)ctr->identity,
                                           count * sizeof(W_));
             break;
+        case HEAP_BY_ERA:
+            fprintf(hp_file, "%lu", (StgWord)ctr->identity);
+            char str_era[100];
+            sprintf(str_era, "%lu", (StgWord)ctr->identity);
+            traceHeapProfSampleString(0, str_era, count * sizeof(W_));
+            break;
         case HEAP_BY_MOD:
         case HEAP_BY_DESCR:
         case HEAP_BY_TYPE:
@@ -1219,6 +1243,10 @@ heapCensusBlock(Census *census, bdescr *bd)
             size = sizeofW(StgTRecChunk);
             break;
 
+        case CONTINUATION:
+            size = continuation_sizeW((StgContinuation *)p);
+            break;
+
         case COMPACT_NFDATA:
             barf("heapCensus, found compact object in the wrong list");
             break;
@@ -1241,13 +1269,123 @@ heapCensusBlock(Census *census, bdescr *bd)
          * use this to scan for the next valid heap closure.
          *
          * Note that not all types of slop are relevant here, only the ones
-         * that can reman after major GC. So essentially just large objects
+         * that can remain after major GC. So essentially just large objects
          * and pinned objects. All other closures will have been packed nice
-         * and thight into fresh blocks.
+         * and tight into fresh blocks.
          */
     }
 }
 
+// determine whether a closure should be assigned to the PRIM cost-centre.
+static bool
+closureIsPrim (StgPtr p)
+{
+  bool prim = false;
+  const StgInfoTable *info = get_itbl((const StgClosure *)p);
+  switch (info->type) {
+    case THUNK:
+    case THUNK_1_1:
+    case THUNK_0_2:
+    case THUNK_2_0:
+    case THUNK_1_0:
+    case THUNK_0_1:
+    case THUNK_SELECTOR:
+    case FUN:
+    case BLACKHOLE:
+    case BLOCKING_QUEUE:
+    case FUN_1_0:
+    case FUN_0_1:
+    case FUN_1_1:
+    case FUN_0_2:
+    case FUN_2_0:
+    case CONSTR:
+    case CONSTR_NOCAF:
+    case CONSTR_1_0:
+    case CONSTR_0_1:
+    case CONSTR_1_1:
+    case CONSTR_0_2:
+    case CONSTR_2_0:
+    case IND:
+    case AP:
+    case PAP:
+    case AP_STACK:
+    case CONTINUATION:
+        prim = false;
+        break;
+
+    case BCO:
+    case MVAR_CLEAN:
+    case MVAR_DIRTY:
+    case TVAR:
+    case WEAK:
+    case PRIM:
+    case MUT_PRIM:
+    case MUT_VAR_CLEAN:
+    case MUT_VAR_DIRTY:
+    case ARR_WORDS:
+    case MUT_ARR_PTRS_CLEAN:
+    case MUT_ARR_PTRS_DIRTY:
+    case MUT_ARR_PTRS_FROZEN_CLEAN:
+    case MUT_ARR_PTRS_FROZEN_DIRTY:
+    case SMALL_MUT_ARR_PTRS_CLEAN:
+    case SMALL_MUT_ARR_PTRS_DIRTY:
+    case SMALL_MUT_ARR_PTRS_FROZEN_CLEAN:
+    case SMALL_MUT_ARR_PTRS_FROZEN_DIRTY:
+    case TSO:
+    case STACK:
+    case TREC_CHUNK:
+        prim = true;
+        break;
+
+    case COMPACT_NFDATA:
+        barf("heapCensus, found compact object in the wrong list");
+        break;
+
+    default:
+        barf("heapCensus, unknown object: %d", info->type);
+  }
+  return prim;
+}
+
+static void
+heapCensusSegment (Census* census, struct NonmovingSegment* seg )
+{
+  unsigned int block_size = nonmovingSegmentBlockSize(seg);
+  unsigned int block_count = nonmovingSegmentBlockCount(seg);
+
+  for (unsigned int b = 0; b < block_count; b++) {
+    StgPtr p = nonmovingSegmentGetBlock(seg, b);
+    // ignore unmarked heap objects
+    if (!nonmovingClosureMarkedThisCycle(p)) continue;
+    // NB: We round up the size of objects to the segment block size.
+    // This aligns with live bytes accounting for the nonmoving collector.
+    heapProfObject(census, (StgClosure*)p, block_size / sizeof(W_), closureIsPrim(p));
+  }
+}
+
+/* Note [Non-concurrent nonmoving collector heap census]
+ * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+ * When using the nonmoving collector, we currently disable concurrent collection
+ * to simplify heap census accounting.
+ *
+ * Without concurrent allocation, marked objects on the nonmoving heap are exactly
+ * the live objects.
+ *
+ * We disable concurrent collection both for GCs that lead to a heap census and not.
+ * This is because a concurrent collection can overlap with a GC that is meant
+ * to perform a heap census. Alternatively we could better handle the case where
+ * a non-concurrent collection is triggered while a non-concurrent collection
+ * is running.
+ */
+
+static void
+heapCensusSegmentList (Census* census, struct NonmovingSegment* seg )
+{
+  for (; seg; seg = seg->link) {
+    heapCensusSegment(census, seg);
+  }
+}
+
 /* -----------------------------------------------------------------------------
  * Code to perform a heap census.
  * -------------------------------------------------------------------------- */
@@ -1310,7 +1448,7 @@ void heapCensus (Time t)
       heapCensusChain( census, generations[g].large_objects );
       heapCensusCompactList ( census, generations[g].compact_objects );
 
-      for (n = 0; n < n_capabilities; n++) {
+      for (n = 0; n < getNumCapabilities(); n++) {
           ws = &gc_threads[n]->gens[g];
           heapCensusChain(census, ws->todo_bd);
           heapCensusChain(census, ws->part_list);
@@ -1318,6 +1456,24 @@ void heapCensus (Time t)
       }
   }
 
+  if (RtsFlags.GcFlags.useNonmoving) {
+    for (unsigned int i = 0; i < nonmoving_alloca_cnt; i++) {
+      heapCensusSegmentList(census, nonmovingHeap.allocators[i].filled);
+      heapCensusSegmentList(census, nonmovingHeap.allocators[i].saved_filled);
+      heapCensusSegmentList(census, nonmovingHeap.allocators[i].active);
+
+      heapCensusChain(census, nonmoving_large_objects);
+      heapCensusCompactList(census, nonmoving_compact_objects);
+
+      // segments living on capabilities
+      for (unsigned int j = 0; j < getNumCapabilities(); j++) {
+        Capability* cap = getCapability(j);
+        heapCensusSegment(census, cap->current_segments[i]);
+      }
+    }
+
+  }
+
   // dump out the census info
 #if defined(PROFILING)
     // We can't generate any info for LDV profiling until
