@@ -25,6 +25,7 @@
 
 #include "rts/PosixSource.h"
 #include "Rts.h"
+#include "RtsFlags.h"
 
 #include "Storage.h"
 #include "GCThread.h"
@@ -401,7 +402,7 @@ void listAllBlocks (ListBlocksCb cb, void *user)
 
       // list capabilities' current segments
       if(RtsFlags.GcFlags.useNonmoving) {
-        for (s = 0; s < NONMOVING_ALLOCA_CNT; s++) {
+        for (s = 0; s < nonmoving_alloca_cnt; s++) {
           listSegmentBlocks(cb, user, getCapability(i)->current_segments[s]);
         }
       }
@@ -409,7 +410,7 @@ void listAllBlocks (ListBlocksCb cb, void *user)
 
   // list blocks on the nonmoving heap
   if(RtsFlags.GcFlags.useNonmoving) {
-    for(s = 0; s < NONMOVING_ALLOCA_CNT; s++) {
+    for(s = 0; s < nonmoving_alloca_cnt; s++) {
       listSegmentBlocks(cb, user, nonmovingHeap.allocators[s].filled);
       listSegmentBlocks(cb, user, nonmovingHeap.allocators[s].saved_filled);
       listSegmentBlocks(cb, user, nonmovingHeap.allocators[s].active);
@@ -1440,7 +1441,7 @@ dirty_MUT_VAR(StgRegTable *reg, StgMutVar *mvar, StgClosure *old)
     Capability *cap = regTableToCapability(reg);
     // No barrier required here as no other heap object fields are read. See
     // Note [Heap memory barriers] in SMP.h.
-    SET_INFO((StgClosure*) mvar, &stg_MUT_VAR_DIRTY_info);
+    SET_INFO_RELAXED((StgClosure*) mvar, &stg_MUT_VAR_DIRTY_info);
     recordClosureMutated(cap, (StgClosure *) mvar);
     IF_NONMOVING_WRITE_BARRIER_ENABLED {
         // See Note [Dirty flags in the non-moving collector] in NonMoving.c
@@ -1462,7 +1463,7 @@ dirty_TVAR(Capability *cap, StgTVar *p,
     // No barrier required here as no other heap object fields are read. See
     // Note [Heap memory barriers] in SMP.h.
     if (RELAXED_LOAD(&p->header.info) == &stg_TVAR_CLEAN_info) {
-        SET_INFO((StgClosure*) p, &stg_TVAR_DIRTY_info);
+        SET_INFO_RELAXED((StgClosure*) p, &stg_TVAR_DIRTY_info);
         recordClosureMutated(cap,(StgClosure*)p);
         IF_NONMOVING_WRITE_BARRIER_ENABLED {
             // See Note [Dirty flags in the non-moving collector] in NonMoving.c
@@ -2005,7 +2006,7 @@ void rts_clearMemory(void) {
             nonmovingClearSegment(seg);
         }
 
-        for (int i = 0; i < NONMOVING_ALLOCA_CNT; ++i) {
+        for (int i = 0; i < nonmoving_alloca_cnt; ++i) {
             struct NonmovingAllocator *alloc = &nonmovingHeap.allocators[i];
 
             for (struct NonmovingSegment *seg = alloc->active; seg; seg = seg->link) {
